import os
from argparse import ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import six
from dodo_commands.framework import ramda as R
from dodo_commands.framework.choice_picker import ChoicePicker
from dodo_commands.framework.command_map import get_command_map
from dodo_commands.framework.command_path import get_command_dirs_from_config
from dodo_commands.framework.config_io import ConfigIO

raw_input = six.moves.input


def _args():
    parser = ArgumentParser(description="Creates a new Dodo command.")
    parser.add_argument("--name")
    parser.add_argument(
        "--yaml", action="store_true", help="Create a yaml command file"
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument("-i", "--interactive", action="store_true")
    group.add_argument("--next-to")

    parser.add_argument(
        "-f", "--force", action="store_true", help="Overwrite existing command script"
    )
    args = Dodo.parse_args(parser)

    if args.next_to and not args.name:
        raise CommandError("The --name argument is mandatory when --next-to is used")

    return args


def create_command_dir(command_dirs):
    def get_command_dir_name():
        while True:
            dir_name = raw_input("\nEnter a name for the commands dir: ")
            if dir_name:
                default_src_dir = os.path.join(Dodo.get("/ROOT/project_dir"), "src")
                src_dir = Dodo.get("/ROOT/src_dir", default_src_dir)
                return os.path.join(src_dir, "extra", "dodo_commands", dir_name)
            else:
                print("Sorry, I did not understand that.")

    def create_dir(new_command_dir):
        if not os.path.exists(new_command_dir):
            Dodo.run(["mkdir", "-p", new_command_dir])

        init_py = os.path.join(new_command_dir, "__init__.py")
        if not os.path.exists(init_py):
            Dodo.run(["touch", init_py])

    def save_config(new_command_dir):
        if not [x for x in command_dirs if os.path.normpath(x) == new_command_dir]:
            config = ConfigIO().load()
            config["ROOT"]["command_path"].append(new_command_dir)
            ConfigIO().save(config)

    command_dir = get_command_dir_name()
    create_dir(command_dir)
    save_config(command_dir)
    return command_dir


def get_command_dir(command_dirs):
    col_width = max(len(os.path.basename(x)) for x in command_dirs)

    def map_to_choice(command_dir):
        return "%s (%s)" % (os.path.basename(command_dir).ljust(col_width), command_dir)

    def get_choice_idx0(choices):
        class Picker(ChoicePicker):
            def print_choices(self, choices):
                for idx, command_dir in enumerate(choices):
                    if idx == len(choices) - 1:
                        print("")
                    print("%d - %s" % (idx + 1, command_dir))
                print()

            def question(self):
                return "Select a target command dir: "

        picker = Picker(choices)
        picker.pick()
        return picker.get_idxs0()[0]

    choices = R.map(map_to_choice)(command_dirs) + ["create a new commands dir"]
    choice_idx0 = get_choice_idx0(choices)
    if choice_idx0 == len(choices) - 1:
        return create_command_dir(command_dirs)
    return command_dirs[choice_idx0]


def get_parser_args(args, builder):
    def get_choice_idxs0(args):
        class Picker(ChoicePicker):
            def print_choices(self, choices):
                print("0 - none")
                for idx, arg in enumerate(choices):
                    print("%d - %s" % (idx + 1, arg[1]))
                print()

            def question(self):
                return "Select the options and args that are given by the user: "

            def on_invalid_index(self, index):
                if index == 0:
                    self.idxs = []

        picker = Picker(args)
        picker.pick()
        return picker.get_idxs0()

    def convert_to_single_string(args, idxs):
        parts = []
        for idx in idxs:
            arg = args[idx]
            parts.append(builder.format_arg_spec(arg[0]))
        return builder.arg_spec_sep.join(parts)

    idxs0 = get_choice_idxs0(args)
    return idxs0, convert_to_single_string(args, idxs0)


def get_params(args, remaining_idxs):
    def get_choices(args):
        class Picker(ChoicePicker):
            def print_choices(self, choices):
                print("0 - none")
                for idx, arg in enumerate(choices):
                    print("%d - %s" % (idx + 1, arg[1]))
                print()

            def question(self):
                return "Select the options and args that come from the config: "

            def on_invalid_index(self, index):
                if index == 0:
                    self.idxs = []

        remaining_choices = [args[remaining_idx] for remaining_idx in remaining_idxs]
        picker = Picker(remaining_choices)
        picker.pick()
        return [remaining_idxs[x] for x in picker.get_idxs0()]

    def convert_to_single_string(args, idxs):
        parts = []
        for idx in idxs:
            name = args[idx][0]
            clean_name = args[idx][2]
            parts.append(
                "        "
                + "ConfigArg('/PATH/TO/{clean_name}', '{name}'),".format(
                    name=name, clean_name=clean_name
                )
            )
        return "\n".join(parts)

    idxs = get_choices(args)
    return idxs, convert_to_single_string(args, idxs)


script_py = """from argparse import ArgumentParser

from dodo_commands import Dodo, ConfigArg, CommandError


def _args():
    parser = ArgumentParser(
        description='{description}'
    )

    # Add arguments to the parser here
{parser_args_str}
    # Parse the arguments.
    args = Dodo.parse_args(parser, config_args=[
{params_str}
    ])

    args.cwd = Dodo.get('/ROOT/project_dir')

    # Raise an error if something is not right
    if False:
        raise CommandError('Oops')

    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()
    Dodo.run([{args_str}], cwd=args.cwd)
"""


script_yaml = """{name}:
  _args: [{parser_args_str}]
  _description: {description}
  run_script:
    args: 
{args_str}
"""


class YamlBuilder:
    arg_spec_sep = ", "

    @staticmethod
    def dest_path(command_dir, name):
        return os.path.join(command_dir, "dodo." + name + ".yaml")

    @staticmethod
    def format_arg_usage(args, parser_arg_idxs, param_arg_idxs):
        args_str_parts = []
        for idx in range(len(args)):
            arg = args[idx]
            if idx in parser_arg_idxs:
                args_str_parts.append("      - ${/_ARGS/" + arg[2] + "}")
            elif idx in param_arg_idxs:
                args_str_parts.append("      - ${/PATH/TO/" + arg[2] + "}")
            else:
                args_str_parts.append("      - %s" % arg[1])
        return "\n".join(args_str_parts)

    @staticmethod
    def format_config_param_usage(param):
        return

    @staticmethod
    def format_arg_spec(arg):
        return arg


class PyBuilder:
    arg_spec_sep = "\n"

    @staticmethod
    def dest_path(command_dir, name):
        return os.path.join(command_dir, name + ".py")

    @staticmethod
    def format_arg_usage(args, parser_arg_idxs, param_arg_idxs):
        args_str_parts = []
        for idx in range(len(args)):
            arg = args[idx]
            if idx in parser_arg_idxs or idx in param_arg_idxs:
                args_str_parts.append("args." + arg)
            else:
                args_str_parts.append("'%s'" % arg[1])
        return ", ".join(args_str_parts)

    @staticmethod
    def format_arg_spec(arg):
        return "    parser.add_argument('{name}')".format(name=arg)


def handle_interactive(parsed_args):
    parsed_args = _args()

    builder = YamlBuilder() if parsed_args.yaml else PyBuilder()

    name = raw_input("Enter a name for the command: ")
    description = raw_input("Enter a description: ")

    def get_args():
        def map_to_split_arg(arg):
            arg_name = arg.split("=")[0]

            clean_name = arg_name
            clean_name = R.cut_prefix(clean_name, "--")
            clean_name = R.cut_prefix(clean_name, "-")

            return (arg_name, arg, clean_name)

        raw_args = raw_input("Enter a command line to run inside the script: ")
        raw_args = raw_args.split(" ")
        return R.map(map_to_split_arg)(raw_args)

    command_dirs = Dodo.get_container().commands.command_dirs
    command_dir = get_command_dir(command_dirs)

    dest_path = builder.dest_path(command_dir, name)

    if os.path.exists(dest_path) and not parsed_args.force:
        raise CommandError("Destination already exists: %s" % dest_path)

    args = get_args()
    parser_arg_idxs, parser_args_str = get_parser_args(args, builder)

    remaining_arg_idxs = [x for x in range(len(args)) if x not in parser_arg_idxs]
    param_arg_idxs, params_str = get_params(args, remaining_arg_idxs)

    args_str = builder.format_arg_usage(args, parser_arg_idxs, param_arg_idxs)

    if parsed_args.yaml:
        with open(dest_path, "w") as f:
            f.write(
                script_yaml.format(
                    name=name,
                    parser_args_str=parser_args_str,
                    params_str=params_str,
                    description=description,
                    args_str=args_str,
                )
            )
    else:
        with open(dest_path, "w") as f:
            f.write(
                script_py.format(
                    parser_args_str=parser_args_str,
                    params_str=params_str,
                    description=description,
                    args_str=args_str,
                )
            )
    print(dest_path)


def handle_next_to(parsed_args):
    command_dirs = get_command_dirs_from_config(Dodo.get())
    command_map = get_command_map(command_dirs)
    item = command_map.get(parsed_args.next_to)

    if not item:
        raise CommandError("Script not found: %s" % parsed_args.next_to)

    dest_path = os.path.join(os.path.dirname(item.filename), parsed_args.name + ".py")

    if os.path.exists(dest_path) and not parsed_args.force:
        raise CommandError("Destination already exists: %s" % dest_path)

    with open(dest_path, "w") as f:
        f.write(
            script_py.format(
                parser_args_str="", params_str="", description="", args_str=""
            )
        )
    print(dest_path)


if Dodo.is_main(__name__, safe=False):
    parsed_args = _args()
    if parsed_args.interactive:
        handle_interactive(parsed_args)
    else:
        handle_next_to(parsed_args)



"""Tests for config.py."""

import os

from dodo_commands.dependencies import yaml_round_trip_dump

import pytest

from ..config import load_config
from ..config_io import ConfigIO


class TestConfigIO:  # noqa
    @pytest.fixture
    def simple_config(self, tmpdir):
        foo_dir = os.path.join(str(tmpdir), "foo")
        return {
            "LAYERS": ["mylayer.yml"],
            "ROOT": {
                "command_path": [os.path.join(foo_dir, "bar")],
                "foo": {"bar": "foobar", "one": "two"},
                "one": 1,
            },
        }

    @pytest.fixture
    def layer(self, tmpdir):
        foo_dir = os.path.join(str(tmpdir), "foo")
        return {
            "ROOT": {
                "command_path": [
                    os.path.join(foo_dir, "foobar"),
                ],
                "foo": {"bar": "barfoo", "three": 4},
                "one": 1.1,
            }
        }

    @pytest.fixture
    def create_config(self, tmpdir, simple_config):
        """Create a config file in the tmpdir."""
        config_io = ConfigIO(str(tmpdir))  # noqa
        config_io.save(simple_config)

    @pytest.fixture
    def create_layer(self, tmpdir, layer):
        """Create a config layer file in the tmpdir."""
        layer_filename = os.path.join(str(tmpdir), "mylayer.yml")
        with open(layer_filename, "w") as f:
            f.write(yaml_round_trip_dump(layer))

    def test_save(self, tmpdir, simple_config):  # noqa
        config_io = ConfigIO(str(tmpdir))  # noqa
        config_io.save(simple_config)
        config = config_io.load()
        assert simple_config == config

    @pytest.mark.usefixtures("create_config", "create_layer")
    def test_layers(self, tmpdir):  # noqa
        foo_dir = os.path.join(str(tmpdir), "foo")
        config_io = ConfigIO(str(tmpdir))  # noqa
        full_config = load_config(["config.yaml", "mylayer.yml"], config_io=config_io)

        assert full_config["ROOT"]["command_path"] == [
            os.path.join(foo_dir, "bar"),
            os.path.join(foo_dir, "foobar"),
        ]
        assert full_config["ROOT"]["foo"] == {
            "bar": "barfoo",
            "one": "two",
            "three": 4,
        }
        assert full_config["ROOT"]["one"] == 1.1



from dodo_commands.framework import ramda as R


# if there are two args that equal "--" then we assume that everything
# behind the second instance of "--" is a part of the "left side" of the
# command line
def rearrange_double_dash(args):
    left, rest = R.split_when(R.equals("--"), args)
    middle, right = [], []
    if rest:
        middle, rest = R.split_when(R.equals("--"), rest[1:])
        if rest:
            right, rest = R.split_when(R.equals("--"), rest[1:])
    return left + right + (R.prepend("--")(middle) if middle else [])



from argparse import ArgumentParser
import os
import glob

from dodo_commands import Dodo


def _args():
    parser = ArgumentParser()
    parser.add_argument("where")
    parser.add_argument("what")
    parser.add_argument("--pattern", default="*")
    parser.add_argument("--replace")
    args = Dodo.parse_args(parser)
    return args


def _replace(where, what, replace_with):
    for filepath in glob.iglob(
        os.path.join(where, "**/" + args.pattern), recursive=True
    ):
        with open(filepath) as file:
            s = file.read()
        s2 = s.replace(what, replace_with)
        if s != s2:
            with open(filepath, "w") as file:
                file.write(s2)


args = _args()
if Dodo.is_main(__name__, safe=not args.replace):
    if args.replace:
        _replace(args.where, args.what, args.replace)
    else:
        Dodo.run(["grep", "-rnw", args.where, "-e", "'{}'".format(args.what)])



import os

from dodo_commands.framework.paths import Paths

from ._create_config_dir import create_config_dir
from ._create_dodo_script import create_dodo_script
from ._create_env_dir import create_env_dir
from ._create_python_env import create_python_env


def _python_env_filename(python_env_dir, basename):
    return os.path.join(python_env_dir, "bin", basename)


def _remove_dir(x):
    if os.path.exists(x):
        print("Clean-up step: rm -rf %s" % x)


def _remove_file(x):
    if os.path.exists(x):
        print("Clean-up step: rm %s" % x)


def register_env(env, project_dir, config_dir, python_env_dir, undo_steps):
    global_bin_dir = os.path.join(Paths().global_config_dir(), "bin")
    named_dodo_filename = os.path.join(global_bin_dir, "dodo-%s" % env)
    env_dir = Paths().env_dir(env)
    python_path = (
        _python_env_filename(python_env_dir, "python") if python_env_dir else None
    )

    if not os.path.exists(global_bin_dir):
        os.makedirs(global_bin_dir)

    if not os.path.exists(project_dir):
        undo_steps.append(lambda: _remove_dir(project_dir))
        os.makedirs(project_dir)

    if not os.path.exists(config_dir):
        undo_steps.append(lambda: _remove_dir(config_dir))
        create_config_dir(config_dir)

    if not os.path.exists(named_dodo_filename):
        undo_steps.append(lambda: _remove_file(named_dodo_filename))
        create_dodo_script(env, named_dodo_filename, python_path)

    undo_steps.append(lambda: _remove_dir(env_dir))
    create_env_dir(env, env_dir, project_dir, config_dir, python_env_dir)


def forget_env(env):
    global_bin_dir = os.path.join(Paths().global_config_dir(), "bin")
    named_dodo_filename = os.path.join(global_bin_dir, "dodo-%s" % env)

    env_dir = Paths().env_dir(env)

    python_env_dir = os.path.join(env_dir, "python_env_dir")
    if os.path.exists(python_env_dir):
        python_env_dir = os.path.realpath(python_env_dir)
        dodo_filename = _python_env_filename(python_env_dir, "dodo")
        _remove_file(dodo_filename)

    _remove_file(named_dodo_filename)
    _remove_dir(env_dir)


def register_python_env(python, env, python_env_dir, undo_steps):
    dodo_filename = _python_env_filename(python_env_dir, "dodo")
    python_path = _python_env_filename(python_env_dir, "python")

    if not os.path.exists(python_env_dir):
        undo_steps.append(lambda: _remove_dir(python_env_dir))
        create_python_env(python, python_env_dir, env)

    if not os.path.exists(dodo_filename):
        undo_steps.append(lambda: _remove_file(dodo_filename))
        create_dodo_script(None, dodo_filename, python_path)



import os
import platform
import shutil
from argparse import ArgumentParser

# Since we'd like the user to install meld, we require it here. If it's not found
# then it will be installed via the meta file (setup.meta).
from dodo_commands.dependencies.get import plumbum, six
from dodo_commands.framework.config import CommandError, Paths
from dodo_commands.framework.singleton import Dodo
from dodo_commands.framework.util import (
    exe_exists,
    is_using_system_dodo,
    query_yes_no,
    symlink,
)

meld = plumbum.cmd.meld  # noqa
raw_input = six.moves.input


def _args():
    parser = ArgumentParser(description="Set up Dodo Commands after installation")

    # Parse the arguments.
    # Optionally, use `config_args` to include additional arguments whose values
    # come either from the configuration file or from the command line (see docs).
    args = Dodo.parse_args(parser)
    return args


def _install_commands(pip_package, confirm):
    if confirm or query_yes_no(
        "Install package %s " % pip_package + "into the default commands directory"
    ):
        Dodo.run(["dodo", "install-commands", "--pip", pip_package])


def _create_left_file(left_dir, filename, contents):
    left_filename = os.path.join(left_dir, filename)
    with open(left_filename, "w") as ofs:
        ofs.write(contents)
    return left_filename


def _create_right_file(right_dir, filename, source_path):
    right_filename = os.path.join(right_dir, filename)
    if os.path.exists(source_path):
        symlink(source_path, right_filename)
    return right_filename


bash_profile = """
# Only in case you've install Dodo Commands in a virtual env: uncomment this line
# and set the path to the directory that holds the system dodo executable.
#
# export PATH=$PATH:/path/to/virtual/env/bin

# Load the .dodo_commands_autostart file
if [ -f ~/.dodo_commands_autostart ]; then
    . ~/.dodo_commands_autostart
fi

# Add a short-cut for $(dodo <command_name>). It let's you use sdodo <command_name>.
sdodo() {
    $(dodo $@)
}
"""

global_config = """[alias]
wh = which
pc = print-config
ec = edit-config
dk = docker-kill
de = docker-exec
dc = docker-commit
mr = menu --run
mm = menu
"""

meld_instructions = """
The setup command will now assist you in comparing your configuration files
to the prototypical ones by comparing files in meld.
Just copy the changes that you agree with from the files on the left side
to the files on the right side.
"""

no_meld_instructions = """
Since meld is not installed, the setup command will not assist you directly
in comparing your configuration files to the prototypical ones. However, you
can do this manually by comparing %s to %s.
Press any key when you are done with the manual comparison, so setup will
continue to remove these temporary directories.
"""

# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()
    if not is_using_system_dodo():
        raise CommandError(
            "Please activate the default environment first by running 'dodo env default'."
        )

    if not args.confirm:
        if query_yes_no("Turn on confirmation for each command that setup executes?"):
            os.environ["__DODO_UNIVERSAL_CONFIRM__"] = "1"

    is_darwin = platform.system() == "Darwin"
    is_linux = platform.system() == "Linux"

    if exe_exists("meld"):
        Dodo.run(["dodo", "global-config", "settings.diff_tool", "meld"])

    Dodo.run(["dodo", "autostart", "on"])

    for pip_package in ("dodo_webdev_commands", "dodo_git_commands"):
        _install_commands(pip_package, args.confirm)

    # tmp_dir = tempfile.mkdtemp()
    tmp_dir = "/tmp/dodo_setup"
    shutil.rmtree(tmp_dir)  # HACK

    left_dir = os.path.join(tmp_dir, "left")
    if not os.path.exists(left_dir):
        os.makedirs(left_dir)
    right_dir = os.path.join(tmp_dir, "right")
    if not os.path.exists(right_dir):
        os.makedirs(right_dir)

    names = dict(bash_profile=".bash_profile" if is_darwin else ".bashrc")

    _create_left_file(left_dir, names["bash_profile"], bash_profile)
    bash_profile_path = os.path.join(os.path.expanduser("~"), names["bash_profile"])
    if not os.path.exists(bash_profile_path):
        print("The bash profile file %s does not exist." % bash_profile_path)
        if query_yes_no("Create it?"):
            Dodo.run(["touch", bash_profile_path])
    _create_right_file(right_dir, names["bash_profile"], bash_profile_path)

    _create_left_file(left_dir, "config", global_config)
    _create_right_file(right_dir, "config", Paths().global_config_filename())

    if not exe_exists("meld"):
        print(no_meld_instructions % (left_dir, right_dir))
        raw_input()
    else:
        print(meld_instructions)
        Dodo.run(["meld", left_dir, right_dir])

    shutil.rmtree(tmp_dir)



import re

from dodo_commands.dependencies.get import parsimonious, six

raw_input = six.moves.input
ParseError = parsimonious.exceptions.ParseError
Grammar = parsimonious.grammar.Grammar
NodeVisitor = parsimonious.nodes.NodeVisitor


def parse_choice(raw_choice):
    grammar = Grammar(
        r"""
      choices   = (term "," choices) / term
      term       = double / single
      single     = number ""
      double     = first "-" last
      first      = number ""
      last       = number ""
      number     = ~r"[\d]+"
      """
    )

    def remove_whitespace(x):
        pattern = re.compile(r"\s+")
        return re.sub(pattern, "", x)

    class Visitor(NodeVisitor):
        def __init__(self):
            self.choices = []

        def visit_single(self, node, visited_children):
            choice = int(node.text)
            self.choices.append((choice, choice))

        def visit_first(self, node, visited_children):
            choice = int(node.text)
            self.choices.append((choice, None))

        def visit_last(self, node, visited_children):
            choice = int(node.text)
            self.choices[-1] = (self.choices[-1][0], choice)

        def generic_visit(self, node, visited_children):
            return visited_children or node

    tree = grammar.parse(remove_whitespace(raw_choice))
    visitor = Visitor()
    visitor.visit(tree)

    idxs = []
    for from_index, to_index in visitor.choices:
        for idx in range(from_index, to_index + 1):
            idxs.append(idx)
    return idxs


class ChoicePicker:
    def __init__(self, choices, start_index=1, allow_free_text=False, labels=None):
        self._choices = choices
        self._labels = labels
        self._start_index = start_index
        self.allow_free_text = allow_free_text
        self.free_text = None

    def print_choices(self):
        raise NotImplementedError()

    def question(self):
        raise NotImplementedError()

    def on_invalid_index(self, index):
        pass

    def pick(self):
        self.free_text = None
        self.idxs = []
        self.print_choices(self._labels or self._choices)
        while True:
            raw_choice = raw_input(self.question())
            try:
                self.idxs = parse_choice(raw_choice)
            except ParseError:
                if self.allow_free_text:
                    self.free_text = raw_choice
                    return
                print("Sorry, I did not understand that")
            else:
                for idx in self.idxs:
                    idx0 = idx - self._start_index
                    if (idx0 < 0) or (idx0 >= len(self._choices)):
                        self.on_invalid_index(idx)
                return

    def get_choices(self):
        return [self._choices[x] for x in self.get_idxs0()]

    def get_idxs0(self):
        return [x - self._start_index for x in self.idxs]



import importlib
import os
import sys
from types import ModuleType


class Switcher:
    def __init__(self):
        self.is_private = False
        self.public_sys_modules = {}
        self.private_sys_modules = {}
        self.package_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "packages"
        )
        self.built_in = {}
        for x in sys.builtin_module_names:
            self.built_in[x] = True

    def switch_to_private(self):
        # Move non-built-in modules to self.public_sys_modules
        for k, v in list(sys.modules.items()):
            if not self.built_in.get(k):
                self.public_sys_modules[k] = v
                del sys.modules[k]

        # Copy private modules to sys.modules
        for k, v in self.private_sys_modules.items():
            sys.modules[k] = v
        self.private_sys_modules.clear()

        sys.path = [self.package_path] + sys.path
        self.is_private = True

    def switch_to_public(self):
        # Move non-built-in modules to self.private_sys_modules
        for k, v in list(sys.modules.items()):
            if not self.built_in.get(k):
                self.private_sys_modules[k] = v
                del sys.modules[k]

        # Copy public modules to sys.modules
        for k, v in self.public_sys_modules.items():
            sys.modules[k] = v
        self.public_sys_modules.clear()

        sys.path.remove(self.package_path)
        self.is_private = False


switcher = Switcher()


class private_sys_modules:
    def __enter__(self):
        self.is_switching = not switcher.is_private
        if self.is_switching:
            switcher.switch_to_private()
        return self

    def __exit__(self, *exc):
        if self.is_switching and switcher.is_private:
            switcher.switch_to_public()
        return False


def get_dependency(path):
    if path in switcher.private_sys_modules:
        return switcher.private_sys_modules[path]

    with private_sys_modules():
        return importlib.import_module(path)


class LocalModule(ModuleType):
    """The module-hack that allows us to use ``from dependencies.get import some_module``"""

    __all__ = ()  # to make help() happy
    __package__ = __name__

    def __getattr__(self, name):
        name_ext = (
            "ruamel.yaml"
            if name == "yaml"
            else "funcy.py2"
            if name == "funcy"
            else name
        )
        return get_dependency(name_ext)

    __path__ = []  # type: List[str]
    __file__ = __file__


get = LocalModule(__name__ + ".get", LocalModule.__doc__)
sys.modules[get.__name__] = get


def yaml_round_trip_load(text):
    with private_sys_modules():
        yaml = get_dependency("ruamel.yaml")
        return yaml.round_trip_load(text)


def yaml_round_trip_dump(text):
    with private_sys_modules():
        yaml = get_dependency("ruamel.yaml")
        return yaml.round_trip_dump(text)



import sys

from dodo_commands.framework import ramda as R
from dodo_commands.framework.command_error import CommandError


def load_named_layers(
    config_io,
    metadata_by_layer_name,
):
    layer_by_target_path = {}

    def get_target_paths_and_layer_names():
        result = {}
        for layer_name, layer_metadata in metadata_by_layer_name.items():
            result[layer_name] = layer_metadata.target_path
        return result.items()

    def do_check_valid_target_path(layer_name, target_path):
        if "*" in target_path:
            raise CommandError("Layer_name may not contain wildcards: %s" % target_path)

        if not config_io.glob([target_path]):
            sys.stderr.write(
                "Warning: layer not found: %s. Check /LAYER_GROUPS for layer %s\n"
                % (target_path, layer_name)
            )

    def add_layer(layer_name, target_path):
        is_found = config_io.glob([target_path])
        layer = config_io.load(target_path) if is_found else {}
        return layer_name, target_path, layer

    def do_store(layer_name, target_path, layer):
        layer_by_target_path[target_path] = layer

    x = get_target_paths_and_layer_names()
    # [path, layer_name]
    x = R.for_each(R.ds(do_check_valid_target_path))(x)
    # [path, layer_name]
    x = R.map(R.ds(add_layer))(x)
    # [(path, layer_name, layer)]
    R.for_each(R.ds(do_store))(x)

    return layer_by_target_path



import os

import dodo_commands
from dodo_commands import CommandError
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.util import symlink


def site_packages_dir(python_env_dir):
    python = plumbum.local[os.path.join(python_env_dir, "bin", "python")]

    result = python(
        "-c",
        "from distutils.sysconfig import get_python_lib; " + "print(get_python_lib())",
    )
    while result[-1] in ["\n", "\r"]:
        result = result[:-1]
    return result


def create_python_env(python, python_env_dir, env):
    try:
        plumbum.local["virtualenv"]("--version")
    except:  # noqa
        raise CommandError("Could not find virtualenv, please install it first.")

    plumbum.local["virtualenv"]("-p", python, python_env_dir, "--prompt", env)

    symlink(
        os.path.realpath(os.path.dirname(dodo_commands.__file__)),
        os.path.join(site_packages_dir(python_env_dir), "dodo_commands"),
    )



import os
from collections import OrderedDict

from dodo_commands.framework import ramda as R


def layer_filename_superset(layer_filenames, config_io):
    selected_layer_by_path = OrderedDict()

    def load_layers(layer_paths):
        def map_to_layer_filenames(layer_paths):
            return config_io.glob(layer_paths)

        def map_to_filename_and_layer(layer_filename):
            return layer_filename, config_io.load(layer_filename)

        def do_store(layer_filename, layer):
            selected_layer_by_path[layer_filename] = layer

        def map_to_nested_layer_paths(layer_filename, layer):
            return R.path_or([], "LAYERS")(layer)

        def get_flat_list(list_of_lists):
            return R.uniq(R.flatten(list_of_lists))

        x = map_to_layer_filenames(layer_paths)
        # [[layer_filename]]
        x = R.map(map_to_filename_and_layer)(x)
        # [layer_filename, layer]
        x = R.for_each(R.ds(do_store))(x)
        # [layer_filename, layer]
        x = R.map(R.ds(map_to_nested_layer_paths))(x)
        # [[layer_filename]]
        x = get_flat_list(x)
        # [layer_filename, layer]
        if x:
            load_layers(x)

    load_layers(layer_filenames)
    return list(selected_layer_by_path.keys())


def get_conflicts_in_layer_paths(layer_paths):
    generic_paths = {}

    def map_to_layer_path_and_parts(path):
        parts = os.path.basename(path).split(".")
        return path, parts

    def has_flavours(path, parts):
        return len(parts) == 3

    def map_to_path_and_generic_path(path, parts):
        generic_path = os.path.join(os.path.dirname(path), parts[0])
        return path, generic_path

    def is_conflict(path, generic_path):
        result = bool(generic_paths.get(generic_path, None))
        if not result:
            generic_paths[generic_path] = path
        return result

    x = layer_paths
    # [path]
    x = R.map(map_to_layer_path_and_parts)(x)
    # [(path, parts)]
    x = R.filter(R.ds(has_flavours))(x)
    # [(path, parts)]
    x = R.map(R.ds(map_to_path_and_generic_path))(x)
    # [(path, generic_path)]
    x = R.filter(R.ds(is_conflict))(x)

    return x



import os

from dodo_commands import Dodo
from dodo_commands.framework.config import Paths


def _args():
    Dodo.parser.add_argument("--alt", help="Run an alternative git command")
    Dodo.parser.add_argument(
        "--message", "-m", dest="message", help="The commit message"
    )
    args = Dodo.parse_args()
    args.cwd = Paths().config_dir()
    return args


if Dodo.is_main(__name__, safe=True):
    args = _args()

    if args.alt:
        Dodo.run(["git", *args.alt.split()], cwd=args.cwd)
    else:
        if not os.path.exists(os.path.join(args.cwd, ".git")):
            Dodo.run(["git", "init"], cwd=args.cwd)

        Dodo.run(["git", "add", "-A"], cwd=args.cwd)
        Dodo.run(
            ["git", "commit", "-m", args.message or "Update configuration"],
            cwd=args.cwd,
        )



import argparse
import os

from dodo_commands.framework import ramda as R
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.container.facets import (
    CommandLine,
    Commands,
    Layers,
    i_,
    o_,
    register,
)
from dodo_commands.framework.container.utils import rearrange_double_dash
from dodo_commands.framework.handle_arg_complete import handle_arg_complete


@register(
    i_(CommandLine, "input_args"),
    o_(CommandLine, "layer_paths_from_input_args"),
    o_(CommandLine, "is_trace"),
    o_(CommandLine, "is_help"),
    o_(CommandLine, "input_args"),
)
def parse_input_args(input_args):
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("-L", "--layer", action="append")
    parser.add_argument("--trace", action="store_true")
    parser.add_argument("--help", action="store_true")

    known_args, args = R.pipe(
        #
        R.always(input_args),
        rearrange_double_dash,
        parser.parse_known_args,
    )(None)

    return dict(
        layer_paths_from_input_args=known_args.layer or [],
        is_trace=known_args.trace,
        is_help=known_args.help,
        input_args=args,
    )


# COMMAND LINE
@register(
    i_(CommandLine, "command_prefix"),
    i_(Layers, "metadata_by_layer_name"),
    o_(CommandLine, "layer_paths_from_command_prefix"),
)
def get_layer_paths_from_command_prefix(
    command_prefix,
    metadata_by_layer_name,
):
    def map_to_path(layer_name):
        if layer_name not in metadata_by_layer_name:
            known_layer_names = metadata_by_layer_name.keys()
            raise CommandError(
                "Unknown layer: %s. Known layers: %s"
                % (layer_name, ", ".join(known_layer_names))
            )
        return metadata_by_layer_name[layer_name].target_path

    layer_names = R.uniq(R.filter(bool)(command_prefix.split(".")))
    return dict(layer_paths_from_command_prefix=R.map(map_to_path)(layer_names))


# COMMAND LINE
@register(
    i_(CommandLine, "command_name"),
    i_(Commands, "layer_name_by_inferred_command"),
    i_(Layers, "metadata_by_layer_name"),
    o_(CommandLine, "layer_paths_inferred_by_command_name"),
)
def get_layer_paths_inferred_by_command_name(
    command_name,
    layer_name_by_inferred_command,
    metadata_by_layer_name,
):
    layer_name = layer_name_by_inferred_command.get(command_name, None)
    layer_metadata = metadata_by_layer_name.get(layer_name)

    return dict(
        layer_paths_inferred_by_command_name=[layer_metadata.target_path]
        if layer_metadata
        else []
    )


# COMMAND LINE
@register(
    i_(CommandLine, "command_name"),
    i_(CommandLine, "input_args"),
    i_(Commands, "command_map"),
    i_(Commands, "aliases", alt_name="command_aliases"),
    i_(Commands, "layer_name_by_inferred_command"),
    i_(Layers, "metadata_by_layer_name"),
    i_(Layers, "layer_by_target_path"),
    o_(CommandLine, "input_args"),
)
def expand_and_autocomplete_command_name(
    command_name,
    input_args,
    command_map,
    command_aliases,
    layer_name_by_inferred_command,
    metadata_by_layer_name,
    layer_by_target_path,
):
    is_autocompleting = "_ARGCOMPLETE" in os.environ
    completed_command_name = (
        handle_arg_complete(
            command_names=list(command_map.keys()),
            inferred_command_names=list(layer_name_by_inferred_command.keys()),
            command_aliases=list(command_aliases.keys()),
            metadata_by_layer_name=metadata_by_layer_name,
            layer_by_target_path=layer_by_target_path,
        )
        if is_autocompleting
        else command_name
    )

    alias_target = command_aliases.get(completed_command_name) or completed_command_name
    new_args = alias_target.split(" ")

    return dict(
        input_args=input_args[:1] + new_args + input_args[2:],
        has_changed=not is_autocompleting and new_args != input_args[1:2],
    )



class ArgsTreeNode:
    def __init__(self, name, args=None, is_horizontal=True):
        self.name = name
        self.children = []
        # this flag controls if nodes are printed horizontally
        self.is_horizontal = is_horizontal
        self.args = args or []

    def add_child(self, node):
        self.children.append(node)
        return node

    def __getitem__(self, name):  # noqa
        return next(x for x in self.children if x.name == name)

    def get(self, name, default_value):
        values = [x for x in self.children if x.name == name]
        return next(values) if values else default_value

    def has_child(self, name):  # noqa
        return len([x for x in self.children if x.name == name])

    def append(self, x):  # noqa
        self.args.append(x)

    def extend(self, x):  # noqa
        self.args.extend(x)

    def flatten(self):
        result = list(self.args)
        for x in self.children:
            result.extend(x.flatten())
        return result

    def to_str(self, indent=0, slash=False):
        terminal = "  \\\n" if slash else "\n"

        if self.is_horizontal:
            if self.args:
                result = indent * " " + " ".join(self.args) + terminal
            else:
                result = ""
        else:
            result = ""
            for x in self.args:
                result += indent * " " + x + terminal
        for child in self.children:
            last_root_child = indent == 0 and child == self.children[-1]
            result += child.to_str(
                indent=indent + 2, slash=slash and not last_root_child
            )
        return result



import os
import shlex
from argparse import ArgumentParser

from dodo_commands.dependencies import yaml_round_trip_load
from dodo_commands.framework.command_map import CommandMapItem
from dodo_commands.framework.config import Key, expand_keys
from dodo_commands.framework.singleton import Dodo
from dodo_commands.framework.util import chop


class YamlCommandMapItem(CommandMapItem):
    def __init__(self, group, filename):
        super(YamlCommandMapItem, self).__init__(group, filename, "yaml")


class YamlCommandHandler:
    def add_commands_to_map(self, command_dirs, file_map, command_map):
        for command_dir, files in file_map.items():
            for file in files:
                command_name, ext = os.path.splitext(os.path.basename(file))
                prefix = "dodo."
                if ext == ".yaml" and command_name.startswith(prefix):
                    command_name = command_name[len(prefix) :]
                    with open(file) as ifs:
                        data = yaml_round_trip_load(ifs.read())
                        for command_name in data.keys():
                            command_map[command_name] = YamlCommandMapItem(
                                group=os.path.basename(command_dir), filename=file
                            )

    def _expand(self, x):
        return os.path.expanduser(os.path.expandvars(expand_keys(Dodo.get(), x)))

    def execute(self, command_map_item, command_name):
        with open(command_map_item.filename) as ifs:
            data = yaml_round_trip_load(ifs.read())

        description = None
        for key, val in data[command_name].items():
            if key == "_description":
                description = val

        parser = ArgumentParser(description=description)

        for key, val in data[command_name].items():
            if key == "_args":
                for arg in val:
                    default_val = None
                    if arg.startswith("--") and "=" in arg:
                        arg, default_val = arg.split("=")
                        parser.add_argument(
                            arg, nargs="?", const=self._expand(default_val)
                        )
                    else:
                        parser.add_argument(
                            arg,
                        )

        args = Dodo.parse_args(parser)
        Dodo.get()["_ARGS"] = args.__dict__

        for name, cmd in data[command_name].items():
            if name.startswith("_"):
                continue

            args = cmd["args"]
            if isinstance(args, str):
                args = shlex.split(args)

            try:
                cwd = self._expand(cmd.get("cwd"))
            except:  # noqa
                cwd = None

            store_xpath = cmd.get("store", "")
            res = Dodo.run(
                [self._expand(x) for x in args], cwd=cwd, capture=bool(store_xpath)
            )
            if store_xpath:
                key = Key(Dodo.get(), store_xpath)
                key.set(chop(res))



from dodo_commands.framework import ramda as R
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.layer_metadata import LayerMetadata


def get_metadata_by_layer_name(root_layer):
    result = {}

    groups = R.path_or({}, "LAYER_GROUPS")(root_layer)
    for group_name, group in groups.items():
        for group_item in group:

            if isinstance(group_item, str):
                layer_name = group_item
                base_name = layer_name
                inferred_commands = []
                target_path = None
            elif isinstance(group_item, dict):
                layer_name, metadata = list(group_item.items())[0]
                metadata = metadata or {}
                inferred_commands = list(metadata.get("inferred_by", []))
                target_path = metadata.get("target_path")
                base_name = metadata.get("base_name", layer_name)

            target_path = target_path or "%s.%s.yaml" % (group_name, base_name)
            layer_metadata = LayerMetadata(target_path, inferred_commands, group_name)

            if layer_name in result:
                prev_layer_metadata = result[layer_name]
                raise CommandError(
                    "Name conlict for layer '%s' in groups '%s' and '%s'"
                    % (
                        layer_name,
                        prev_layer_metadata.group_name,
                        layer_metadata.group_name,
                    )
                )

            result[layer_name] = layer_metadata

    return result



import glob
import hashlib
import json
import os
import sys

from dodo_commands.framework.paths import Paths
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.util import symlink


def _glob_command_dirs(patterns):
    result = []
    for pattern in patterns:
        for x in glob.glob(os.path.expanduser(pattern)):
            if os.path.isdir(x) and not os.path.basename(x).startswith("_"):
                result.append(x)
    return result


def get_command_dirs(include_patterns, exclude_patterns):
    excluded_command_dirs = _glob_command_dirs(exclude_patterns)

    items = sorted(
        [
            x
            for x in _glob_command_dirs(include_patterns)
            if x not in excluded_command_dirs
        ]
    )

    basenames = [os.path.basename(x) for x in items]
    for basename in basenames:
        if basenames.count(basename) > 1:
            raise CommandError("More than 1 command path with name %s" % basename)
    return items


def _create_search_path_dir(command_dirs):
    hash_code = hashlib.md5(json.dumps(command_dirs).encode("utf-8")).hexdigest()
    search_path_dir = os.path.join(
        Paths().global_config_dir(), "search_path", hash_code
    )

    if not os.path.exists(search_path_dir):
        os.makedirs(search_path_dir)
        open(os.path.join(search_path_dir, "__init__.py"), "a").close()
        for command_dir in command_dirs:
            basename = os.path.basename(command_dir)
            target_path = os.path.join(search_path_dir, basename)
            if not os.path.exists(target_path):
                symlink(command_dir, target_path)

    return search_path_dir


def extend_sys_path(command_dirs):
    search_path_dir = _create_search_path_dir(command_dirs)
    if search_path_dir not in sys.path:
        sys.path.append(search_path_dir)


def get_command_dirs_from_config(config):
    root_config = config.get("ROOT", {})
    command_path = root_config.get("command_path", [])
    command_path_exclude = root_config.get("command_path_exclude", [])
    return get_command_dirs(
        include_patterns=command_path, exclude_patterns=command_path_exclude
    )



import os
import sys

from dodo_commands.framework import ramda as R


class CommandLine:
    def __init__(self):
        self.input_args = (
            [sys.executable] if self.is_running_directly_from_script else []
        ) + list(sys.argv)
        self.is_trace = False
        self.is_help = False
        self.layer_paths_inferred_by_command_name = []
        self.layer_paths_from_command_prefix = []
        self.layer_paths_from_input_args = []

    @property
    def is_running_directly_from_script(self):
        return "__DODO__" not in os.environ

    @property
    def layer_paths(self):
        return R.uniq(
            self.layer_paths_inferred_by_command_name
            + self.layer_paths_from_command_prefix
            + self.layer_paths_from_input_args
        )

    @property
    def _split_input1(self):
        input1 = self.input_args[1] if len(self.input_args) > 1 else ""

        # When running directly from a script, then we know the command was not
        # prefixed with layer names.
        if self.is_running_directly_from_script:
            return "", R.cut_suffix(input1, ".py")

        return R.split(input1.rfind(".") + 1)(input1)

    @property
    def command_name(self):
        return self._split_input1[1] or "help"

    @property
    def command_prefix(self):
        return self._split_input1[0]

    def get_trace(self):
        args = [x for x in self.input_args if x != "--trace"]
        args[1] = self.command_name
        args.extend(["--layer=%s" % x for x in self.layer_paths])
        return args

    @staticmethod
    def get(ctr):
        return ctr.command_line



from dodo_commands import Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.config import expand_keys


def _args():
    Dodo.parser.description = "Run docker compose"

    Dodo.parser.add_argument("compose_args", nargs="*")
    Dodo.parser.add_argument("--detach", action="store_true")
    args = Dodo.parse_args()
    args.cwd = Dodo.get("/DOCKER_COMPOSE/cwd")
    args.files = Dodo.get("/DOCKER_COMPOSE/files", None)
    args.map = Dodo.get("/DOCKER_COMPOSE/map", {})
    args.compose_project_name = Dodo.get(
        "/DOCKER_COMPOSE/compose_project_name", Dodo.get("/ROOT/env_name")
    )
    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()

    for src, dest in args.map.items():
        with open(src) as ifs:
            content = expand_keys(Dodo.get(), ifs.read())
        with open(dest, "w") as ofs:
            ofs.write(content)

    def get_file_args():
        result = []
        for f in args.files:
            result.extend(["--file", f])
        return result

    file_args = get_file_args() if args.files else []
    detach_args = ["--detach"] if args.detach else []
    with plumbum.local.env(COMPOSE_PROJECT_NAME=args.compose_project_name):
        Dodo.run(
            ["docker-compose", *file_args, *args.compose_args, *detach_args],
            cwd=args.cwd,
        )



from dodo_commands.framework.singleton import Dodo


# Resp: add the current command_name
# to the list of commands decorated by decorator_name.
class DecoratorScope:
    def __init__(self, decorator_name, remove=False):
        self.decorators = (
            Dodo.get("/ROOT")
            .setdefault("decorators", {})
            .setdefault(decorator_name, [])
        )
        self.prefix = "!" if remove else ""

    def __enter__(self):  # noqa
        self.decorators.append(self.prefix + Dodo.command_name)

    def __exit__(self, type, value, traceback):  # noqa
        self.decorators.remove(self.prefix + Dodo.command_name)



import os
import sys

from dodo_commands import Dodo
from dodo_commands.framework.command_map import execute_script, get_command_map
from dodo_commands.framework.command_path import get_command_dirs


def _args():
    Dodo.parser.description = "Download a command script and run it"

    Dodo.parser.add_argument("url")
    Dodo.parser.add_argument("script_args", nargs="*")

    args = Dodo.parse_args()
    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()
    tmp_dir = os.path.join(os.path.expanduser("~"), ".dodo_commands", "tmp")
    package_dir = os.path.join(tmp_dir, "exec_scripts")

    if ".commands.yaml" in args.url:
        last_slash = args.url.rfind("/")
        url = args.url[:last_slash]
        cmd_name = args.url[last_slash + 1 :]
    else:
        url = args.url
        cmd_name = os.path.splitext(os.path.basename(url))[0]

    Dodo.run(["mkdir", "-p", package_dir])
    Dodo.run(["touch", os.path.join(package_dir, "__init__.py")])
    Dodo.run(["wget", url, "-O", os.path.join(package_dir, os.path.basename(url))])

    command_map = get_command_map(get_command_dirs([package_dir], []))
    sys.argv = sys.argv[1:]
    execute_script(command_map, cmd_name)



import os
import re

from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.config_key import Key
from dodo_commands.framework.util import EnvironMemo, xpath_to_string


class DictKey:
    def __init__(self, the_dict, the_key, xpath):
        self.dict = the_dict
        self.key = the_key
        self.xpath = xpath

    def __repr__(self):  # noqa
        return "DK[%s]" % self.key

    def warning_msg_not_expanded(self):
        return "Unexpanded key {key} at location {xpath}".format(
            key=self.key, xpath=xpath_to_string(self.xpath)
        )

    def create_dict_val(self, expanded_key):  # noqa
        if expanded_key != self.key:
            self.dict[expanded_key] = self.dict[self.key]
            del self.dict[self.key]
        return DictVal(self.dict, expanded_key, self.xpath)


class DictVal:
    def __init__(self, the_dict, the_key, xpath):
        self.dict = the_dict
        self.key = the_key
        self.xpath = xpath

    def get_value(self):
        return self.dict[self.key]

    def replace_value(self, new_value):
        self.dict[self.key] = new_value

    def warning_msg_not_expanded(self):
        return "Unexpanded value {val} at location {xpath}".format(
            val=self.get_value(), xpath=xpath_to_string(self.xpath)
        )

    def __repr__(self):  # noqa
        return "DV[%s]" % self.key


class ListVal:
    def __init__(self, the_list, the_idx, xpath):
        self.idx = the_idx
        self.list = the_list
        self.xpath = xpath

    def get_value(self):
        return self.list[self.idx]

    def replace_value(self, new_value):
        self.list[self.idx] = new_value

    def warning_msg_not_expanded(self):
        return "Unexpanded value {val} at location {xpath}".format(
            val=self.get_value(), xpath=xpath_to_string(self.xpath)
        )

    def __repr__(self):  # noqa
        return "L[%s]" % self.idx


def get_key_expressions(search_string, key_regexp=r"\$\{(/[^\}]*)\}"):
    return [x for x in re.finditer(key_regexp, search_string)]


class ConfigExpander:
    """Expand environment variables and references in the config."""

    def __init__(self, extra_vars=None):
        self.extra_vars = extra_vars or {}

    def _expand_str(self, current_str):
        expanded_str = current_str
        key_expressions = get_key_expressions(current_str)
        known_strs = []

        while key_expressions:
            changed = False
            for key_expression in reversed(key_expressions):
                xpath_string = key_expression.group(1)

                key = Key(self.config, xpath_string)
                if key.exists():
                    expanded_str = (
                        expanded_str[: key_expression.start()]
                        + str(key.get())
                        + expanded_str[key_expression.end() :]
                    )

                    if expanded_str not in known_strs:
                        known_strs.append(expanded_str)
                        changed = True

            if not changed:
                return None

            current_str = expanded_str
            key_expressions = get_key_expressions(current_str)

        with EnvironMemo(self.extra_vars):
            expanded_str = os.path.expandvars(expanded_str)

        env_key_expressions = get_key_expressions(expanded_str, r"\$\{([^\}]*)\}")
        return expanded_str if not env_key_expressions else None

    def _expand(self, raw_obj):
        if isinstance(raw_obj, str):
            return self._expand_str(raw_obj)

        if isinstance(raw_obj, list):
            return [self._expand(x) for x in raw_obj]

        if isinstance(raw_obj, dict):
            return {self._expand(k): self._expand(v) for k, v in raw_obj.items()}

        return raw_obj

    def _schedule_children(self, obj, todo, xpath):
        if isinstance(obj, dict):
            for k in obj:
                todo.append(DictKey(obj, k, xpath + [k]))
            return True
        if isinstance(obj, list):
            for idx in range(len(obj)):
                todo.append(ListVal(obj, idx, xpath + [idx]))
            return True

    def run(self, config, callbacks=None):  # noqa
        nodes = []
        self.config = config
        self._schedule_children(config, nodes, [])

        changed = True
        while len(nodes) and changed:
            changed = False
            new_nodes = list()
            for node in nodes:
                if isinstance(node, DictKey):
                    expanded_key = self._expand(node.key)
                    if expanded_key is None:
                        new_nodes.append(node)
                    else:
                        changed = True
                        new_nodes.append(node.create_dict_val(expanded_key))
                elif isinstance(node, DictVal) or isinstance(node, ListVal):
                    value = node.get_value()
                    if self._schedule_children(value, new_nodes, node.xpath):
                        changed = True
                    else:
                        expanded_value = self._expand(value)
                        if expanded_value is None:
                            new_nodes.append(node)
                        else:
                            changed = True

                            node.replace_value(expanded_value)
                            if xpath_to_string(node.xpath) in (callbacks or {}):
                                callbacks[xpath_to_string(node.xpath)](expanded_value)
                else:
                    raise CommandError("Should not reach here")

            nodes = new_nodes

        warnings = ["Warning: %s\n" % node.warning_msg_not_expanded() for node in nodes]
        return warnings



import os
import sys
import tempfile

from dodo_commands import CommandError, Dodo
from dodo_commands.framework.global_config import load_global_config_parser
from dodo_commands.framework.paths import Paths
from dodo_commands.framework.util import bordered, is_using_system_dodo, symlink


def _args():
    parser = Dodo.parser

    parser.description = (
        "Install command packages into the global "
        + "commands directory. "
        + _packages_in_extra_dir()
    )
    parser.add_argument(
        "paths", nargs="*", help="Create symlinks to these command directories"
    )
    parser.add_argument(
        "--as",
        dest="as_directory",
        help="Use this name for the target commands directory",
    )
    parser.add_argument(
        "--pip", nargs="*", help="Pip install the commands in these packages"
    )
    parser.add_argument(
        "--git", nargs="*", help="Clone a git repo into the commands directory"
    )
    parser.add_argument(
        "--remove",
        action="store_true",
        help="Remove commands from the commands directory",
    )
    parser.add_argument(
        "--to-defaults",
        action="store_true",
        help="Install into the default commands directory",
    )

    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        "--make-default",
        help="Create a symlink to a global commands package in the default commands directory",
    )
    group.add_argument(
        "--remove-default",
        help="Remove a symlink to a global commands package from the default commands directory",
    )

    return Dodo.parse_args()


def check_setuptools():
    if not Dodo.run([_python_path(), "-c", "import setuptools"]):
        _report_error(
            "\n"
            + bordered(
                "Error: your python version does not have setuptools installed.\n"
                + "Check the settings.python option in %s"
                % Paths().global_config_filename()
            )
        )
        sys.exit(1)


def _packages_in_extra_dir():
    extra_dir = Paths().extra_dir()
    packages = [
        x
        for x in os.listdir(extra_dir)
        if os.path.isdir(os.path.join(extra_dir, x)) and not x.startswith("__")
    ]

    if len(packages) == 0:
        return ""

    if len(packages) == 1:
        msg = " The %s package is found automagically " % packages[0]
    else:
        packages[-1] = "and " + packages[-1]
        msg = " The %s packages are found automagically " % ", ".join(packages)

    return (
        msg
        + " in the dodo_commands.extra package"
        + ", e.g. the following works: dodo install-commands %s." % packages[0]
    )


def _report_error(msg):
    sys.stderr.write(msg + os.linesep)


def _remove_package(package, only_from_defaults=False):
    """Install the dir with the default commands."""
    if not only_from_defaults:
        dest_dir = os.path.join(Paths().global_commands_dir(), package)
        if not os.path.exists(dest_dir):
            raise CommandError("Not installed: %s" % dest_dir)
        Dodo.run(["rm", "-rf", dest_dir])

    defaults_dest_dir = os.path.join(Paths().default_commands_dir(), package)
    if os.path.exists(defaults_dest_dir):
        Dodo.run(["rm", defaults_dest_dir])


def _install_package(package, as_directory, install_commands_function, add_to_defaults):
    """Install the dir with the global commands."""
    package_dirname = as_directory or package
    dest_dir = os.path.join(Paths().global_commands_dir(), package_dirname)
    defaults_dest_dir = os.path.join(Paths().default_commands_dir(), package_dirname)

    if add_to_defaults and os.path.exists(defaults_dest_dir):
        _report_error("Error: already installed: %s" % defaults_dest_dir)
        return False

    if not install_commands_function(dest_dir):
        return False

    if add_to_defaults:
        if not os.path.exists(dest_dir):
            _report_error("Error: not found: %s" % dest_dir)
            return False

        if os.name == "nt" and not args.confirm:
            symlink(dest_dir, defaults_dest_dir)
        else:
            Dodo.run(["ln", "-s", dest_dir, defaults_dest_dir])

    return True


def _install_commands_from_path(path, dest_dir, mv=False):
    """Install the dir with the global commands."""
    if not os.path.exists(path):
        alt_path = os.path.join(Paths().extra_dir(), path)
        if os.path.exists(alt_path):
            path = alt_path
        else:
            _report_error("Error: path not found: %s" % path)
            return False

    if os.path.exists(dest_dir):
        _report_error("Error: already installed: %s" % dest_dir)
        return False

    if mv:
        Dodo.run(["mv", path, dest_dir])
    else:
        try:
            if os.name == "nt" and not args.confirm:
                symlink(os.path.abspath(path), dest_dir)
            else:
                Dodo.run(["ln", "-s", os.path.abspath(path), dest_dir])
        except:
            _report_error("Error: could not create a symlink in %s." % dest_dir)

    return True


def _python_path():
    config = load_global_config_parser()
    return config.get("settings", "python_interpreter")


def _install_commands_from_package(package):
    Dodo.run(
        [
            _python_path(),
            "-m",
            "pip",
            "install",
            "--upgrade",
            "--target",
            Paths().global_commands_dir(),
            package,
        ]
    )

    return True


def _clone_git_repo(repo_path):
    tmp_dir = tempfile.mkdtemp()
    Dodo.run(["git", "clone", repo_path], cwd=tmp_dir)
    package = os.listdir(tmp_dir)[0]
    return tmp_dir, package


if Dodo.is_main(__name__):
    args = _args()
    if args.pip and not is_using_system_dodo():
        raise CommandError(
            "Please activate the default environment first by running 'dodo env default'."
        )

    if args.make_default:
        _install_package(args.make_default, None, lambda: True, True)
        sys.exit(0)

    if args.remove_default:
        _remove_package(args.remove_default, only_from_defaults=True)
        sys.exit(0)

    if args.paths:
        for path in args.paths:
            package = os.path.basename(path)
            if args.remove:
                _remove_package(package)
            else:
                _install_package(
                    package,
                    args.as_directory,
                    lambda dest_dir: _install_commands_from_path(path, dest_dir),
                    args.to_defaults,
                )

    if args.pip:
        check_setuptools()
        for package in args.pip:
            if args.remove:
                _remove_package(package)
            else:
                if args.as_directory:
                    raise CommandError(
                        "Sorry, the --as option is not supported when --pip is used."
                    )
                _install_package(
                    package,
                    args.as_directory,
                    lambda dest_dir: _install_commands_from_package(package),
                    args.to_defaults,
                )

    if args.git:
        for repo_path in args.git:
            if args.remove:
                raise CommandError(
                    "The --git option is not supported when removing a package."
                    + " Please use dodo install-commands --remove <package>."
                )

            tmp_dir, package = _clone_git_repo(repo_path)
            _install_package(
                package,
                args.as_directory,
                lambda dest_dir: _install_commands_from_path(
                    os.path.join(tmp_dir, package), dest_dir, mv=True
                ),
                args.to_defaults,
            )
            Dodo.run(["rm", "-rf", tmp_dir])



import os
import sys

from dodo_commands.framework.command_error import CommandError  # noqa
from dodo_commands.framework.command_map import execute_script
from dodo_commands.framework.singleton import Dodo  # noqa
from dodo_commands.framework.version import get_version


def _handle_exception(e):
    if "_ARGCOMPLETE" in os.environ:
        sys.exit(0)

    if "--traceback" in sys.argv or not isinstance(e, CommandError):
        raise
    sys.stderr.write("%s: %s\n" % (e.__class__.__name__, e))
    sys.exit(1)


def execute_from_command_line(argv):
    os.environ["__DODO__"] = "1"

    if "_ARGCOMPLETE" in os.environ:
        sys.argv += os.environ["COMP_LINE"].split()[1:2]

    try:
        ctr = Dodo.get_container()

        # TODO: move to action_execute_command
        if ctr.command_line.command_name == "--version":
            print(get_version())
            sys.exit(0)

        if ctr.command_line.is_trace:
            sys.stderr.write("%s\n" % ctr.command_line.get_trace())
            sys.exit(0)

        if ctr.command_line.command_name not in ctr.commands.command_map:
            raise CommandError(
                "Unknown dodo command: %s" % ctr.command_line.command_name
            )

        execute_script(ctr.commands.command_map, ctr.command_line.command_name)
    except KeyboardInterrupt:
        print("\n")
        sys.exit(1)
    except Exception as e:
        _handle_exception(e)



import glob
import os
import sys
from argparse import ArgumentParser

from dodo_commands import Dodo
from dodo_commands.dependencies import yaml_round_trip_load
from dodo_commands.framework.command_path import get_command_dirs_from_config
from dodo_commands.framework.decorator_utils import _all_decorators
from dodo_commands.framework.paths import Paths


def _args():  # noqa
    parser = ArgumentParser(
        description=(
            "Print the value of /ROOT/<what>_dir. For example: "
            + '"dodo which src" prints the value of /ROOT/src_dir.'
        )
    )
    parser.add_argument(
        "what",
        nargs="?",
    )
    parser.add_argument(
        "--cd", action="store_true", help="Prefix the output with 'cd '"
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--config", action="store_true", help="Print where the config file is"
    )
    group.add_argument(
        "--global-config",
        action="store_true",
        help="Print where the global config file is",
    )
    group.add_argument(
        "--env-dir",
        action="store_true",
        help="Print where the environment directory is",
    )
    group.add_argument(
        "--project-dir",
        action="store_true",
        help="Print where the project directory is",
    )
    group.add_argument(
        "--config-dir", action="store_true", help="Print where the config directory is"
    )
    group.add_argument(
        "--python-env-dir",
        action="store_true",
        help="Print where the python virtual env directory is",
    )
    group.add_argument(
        "--script", help="Print where the dodo command script with given name is"
    )
    group.add_argument(
        "--decorators",
        action="store_true",
        help="Prints which command decorators are available",
    )
    group.add_argument(
        "--envs", action="store_true", help="Prints which environments are available"
    )
    group.add_argument(
        "--env", action="store_true", help="Prints name of the active environment"
    )
    group.add_argument(
        "--layers", action="store_true", help="Prints which layers are active"
    )
    group.add_argument(
        "--default-commands",
        action="store_true",
        help="Prints the directory where default commands are stored",
    )
    group.add_argument(
        "--global-commands",
        action="store_true",
        help="Prints the directory where global commands are stored",
    )

    return Dodo.parse_args(parser)


def _which_script(script):
    command_dirs = get_command_dirs_from_config(Dodo.get())
    for item in command_dirs:
        script_path = os.path.join(item, script + ".py")
        if os.path.exists(script_path):
            return os.path.realpath(script_path)

    for item in command_dirs:
        for yml in glob.glob(os.path.join(item, "dodo.*.yaml")):
            with open(yml) as ifs:
                if script in yaml_round_trip_load(ifs.read()).keys():
                    return os.path.realpath(yml)

    return None


def _which_dir(directory):
    return Dodo.get("/ROOT/%s_dir" % directory, None)


if Dodo.is_main(__name__):
    args = _args()

    def report(x):
        prefix = "cd " if args.cd else ""
        sys.stdout.write("%s%s" % (prefix, x))

    if args.config:
        if os.path.exists(Paths().config_dir()):
            report(os.path.join(Paths().config_dir(), "config.yaml"))
    elif args.global_config:
        report(Paths().global_config_filename() + "\n")
    elif args.global_commands:
        report(Paths().global_commands_dir() + "\n")
    elif args.default_commands:
        report(Paths().default_commands_dir() + "\n")
    elif args.script:
        report(_which_script(args.script) + "\n")
    elif args.env_dir:
        report(Paths().env_dir() + "\n")
    elif args.project_dir:
        report(_which_dir("project") + "\n")
    elif args.config_dir:
        report(_which_dir("config") + "\n")
    elif args.python_env_dir:
        python_env_dir = os.path.realpath(
            os.path.join(Paths().env_dir(), "python_env_dir")
        )
        if os.path.exists(python_env_dir):
            report(python_env_dir + "\n")
    elif args.decorators:
        report(", ".join(sorted(_all_decorators(Dodo.get()).keys())) + "\n")
    elif args.envs:
        report("\n".join(sorted(os.listdir(Paths().envs_dir()))) + "\n")
    elif args.env:
        if Dodo.get("/ROOT/env_name", None):
            report(Dodo.get("/ROOT/env_name") + "\n")
    elif args.layers:
        layer_paths = Dodo.get_container().layers.get_ordered_layer_paths()
        for layer_path in layer_paths:
            report(layer_path + "\n")
    elif args.what:
        x = _which_script(args.what)
        if x:
            report(x + "\n")
    else:
        if Dodo.get("/ROOT/env_name", None):
            report(Dodo.get("/ROOT/env_name") + "\n")



import os

from dodo_commands import Dodo


def _args():
    Dodo.parser.add_argument("number", nargs="?")
    Dodo.parser.add_argument("--list", action="store_true")
    Dodo.parser.add_argument("--group", "-g", default="default")
    args = Dodo.parse_args()
    args.dirs = Dodo.get("/DIAL", {})
    return args


if Dodo.is_main(__name__):
    args = _args()

    if args.list:
        Dodo.run(["dodo", "print-config", "DIAL"])
    else:
        dirs = args.dirs.get(args.group, {}) if args.group else args.dirs

        for k, v in dirs.items():
            if args.number == str(k):
                dial_dir = os.path.expanduser(os.path.expandvars(v))
                print(dial_dir)



from argparse import ArgumentParser

from dodo_commands import Dodo


def _args():  # noqa
    parser = ArgumentParser(description=("Print hello world."))
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__):
    args = _args()
    Dodo.run(["echo", "Hello world"])



from dodo_commands import Dodo
from dodo_commands.framework.args_tree import ArgsTreeNode
from dodo_commands.framework.decorator_utils import uses_decorator


class Decorator:  # noqa
    def add_arguments(self, parser):  # noqa
        parser.add_argument(
            "--use-debugger",
            action="store_true",
            default=False,
            help="Run the command through the debugger",
        )

    def is_used(self, config, command_name, decorator_name):
        return uses_decorator(config, command_name, decorator_name)

    def modify_args(self, command_line_args, root_node, cwd):  # noqa
        if not getattr(command_line_args, "use_debugger", False):
            return root_node, cwd

        debugger_node = ArgsTreeNode("debugger", args=[Dodo.get("/BUILD/debugger")])
        debugger_node.add_child(root_node)
        return debugger_node, cwd



import os
from argparse import ArgumentParser

from dodo_commands.dependencies.get import argcomplete
from dodo_commands.framework import ramda as R
from dodo_commands.framework.config_layers import get_conflicts_in_layer_paths
from dodo_commands.framework.get_aliases import get_aliases


def handle_arg_complete(
    command_names,
    inferred_command_names,
    command_aliases,
    metadata_by_layer_name,
    layer_by_target_path,
):
    def get_args():
        return os.environ["COMP_LINE"].split()

    def get_prefix_and_command(args):
        return R.split(args[1].rfind(".") + 1)(args[1])

    input_args = get_args()
    full_command_name = input_args[1]
    # [arg]
    (command_prefix, command_name) = get_prefix_and_command(input_args)

    def get_used_layer_names(command_prefix):
        return [x for x in command_prefix.split(".") if x in metadata_by_layer_name]

    used_layer_names = get_used_layer_names(command_prefix)

    def get_commands():
        more_aliases = []
        for layer_name in used_layer_names:
            layer_metadata = metadata_by_layer_name[layer_name]
            layer = layer_by_target_path[layer_metadata.target_path]
            more_aliases.extend(get_aliases(layer).keys())

        return R.uniq(
            command_names + inferred_command_names + command_aliases + more_aliases
        )

    def get_choices(commands):
        return [(command_prefix + x) for x in commands]

    commands = get_commands()  # [command_name]
    choices = get_choices(commands)  # [choice]

    def get_possible_layer_names():
        return metadata_by_layer_name.keys()

    def is_already_used(layer_name):
        return layer_name in used_layer_names

    def is_conflicting(layer_name):
        used_layer_metadata = R.map(metadata_by_layer_name)(used_layer_names)
        used_layer_paths = [x.target_path for x in used_layer_metadata]

        layer_path = metadata_by_layer_name[layer_name].target_path
        return get_conflicts_in_layer_paths(used_layer_paths + [layer_path])

    def map_to_choices(layer_name):
        return [command_prefix + layer_name + "." + x for x in commands]

    x = get_possible_layer_names()
    # [layer_name]
    x = R.remove_if(is_already_used)(x)
    # [layer_name]
    x = R.remove_if(is_conflicting)(x)
    # [layer_name]
    new_choices = R.map(map_to_choices)(x)
    # [[new_choice]]
    choices += R.flatten(new_choices)

    if full_command_name not in choices:
        parser = ArgumentParser()
        parser.add_argument("command", choices=choices)
        argcomplete.autocomplete(parser)

    os.environ["COMP_LINE"] = " ".join(input_args[:1] + input_args[2:])
    os.environ["COMP_POINT"] = str(
        int(os.environ["COMP_POINT"]) - (len(full_command_name) + 1)
    )

    return command_name



"""Tests for config.py."""

from ..config import ConfigExpander


class TestConfigExpander:  # noqa
    simple_config = {
        "ROOT": {"command_path": ["foo", "bar"]},
        "FOO": {"bar": {"one": "two"}, "one": [1, 2, 3], "two": 2.2},
    }

    config_with_refs = {
        "ROOT": {"command_path": ["foo $PATH", "bar"]},
        "FOO": {
            "bar": {"one": "${/ROOT/command_path/1} and ${/FOO/two}"},
            "one": [1, 2, 3],
            "two": "${/FOO/one/2}",
            "three": "${/FOO/one}",
        },
    }

    def test_config_expander(self):
        """Config values are converted into environment variables."""
        config_expander = ConfigExpander()
        config_expander.run(self.simple_config)

    def test_expand(self):
        """dodo/system environment variables are expanded."""
        config_expander = ConfigExpander()
        config_expander.run(self.config_with_refs)
        assert self.config_with_refs["FOO"]["bar"]["one"] == "bar and 3"
        assert self.config_with_refs["FOO"]["three"] == "[1, 2, 3]"
        assert "$PATH" not in self.config_with_refs["ROOT"]["command_path"][0]



from argparse import ArgumentParser

from dodo_commands import Dodo, CommandError
from dodo_commands.framework.global_config import (
    load_global_config_parser,
    write_global_config_parser,
)


def _args():
    parser = ArgumentParser(
        description="Write a value of the global Dodo Command configuration"
    )
    parser.add_argument("key")
    parser.add_argument("val", nargs="?")
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__, safe=False):
    args = _args()
    config = load_global_config_parser()

    if args.key.count(".") != 1:
        raise CommandError("The key should have the format <section>.<value>")

    section, key = args.key.split(".")
    if args.val:
        config.set(section, key, args.val)
        write_global_config_parser(config)
    else:
        print(config.get(section, key))



import sys

from dodo_commands.framework.container import actions, facets
from dodo_commands.framework.container.facets import run
from dodo_commands.framework.paths import Paths


class Container:
    def _create_facets(self):
        self.paths = Paths()
        self.command_line = facets.CommandLine()
        self.layers = facets.init_layers(facets.Layers(), "config.yaml")
        self.config = facets.Config()
        self.commands = facets.init_commands(facets.Commands())

    def __init__(self):
        self._create_facets()

    @staticmethod
    def get(ctr):
        return ctr

    def run_actions(self):
        run(self, actions.layers.load_root_layer)
        run(self, actions.layers.get_metadata_by_layer_name)
        run(self, actions.layers.load_named_layers)
        run(self, actions.commands.get_inferred_command_map)
        run(self, actions.command_line.get_layer_paths_from_command_prefix)
        run(self, actions.command_line.parse_input_args)

        while True:
            run(self, actions.command_line.get_layer_paths_inferred_by_command_name)

            # Create the config
            run(self, actions.layers.select_layers)
            run(self, actions.config.check_conflicts_in_selected_layer_paths)
            run(self, actions.config.build_from_selected_layers)

            run(self, actions.commands.get_aliases_from_config)
            run(self, actions.commands.get_command_map)

            if not run(self, actions.command_line.expand_and_autocomplete_command_name)[
                "has_changed"
            ]:
                break

        if self.config.warnings:
            for warning in self.config.warnings:
                sys.stderr.write(warning)



import os
import sys

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.util import query_yes_no

local = plumbum.local


def _count_confirm_in_argv():
    result = 0
    for arg in sys.argv:
        if arg == "--confirm":
            result += 1
        if arg.startswith("-") and not arg.startswith("--"):
            result += arg.count("c")
        if arg == "--":
            break
    return result


def _ask_confirmation(args, cwd, is_echo, is_confirm):
    """Ask the user whether to continue with executing func."""

    def to_str():
        return args.to_str(slash=args.children)

    if is_echo:
        print(to_str())
        return False

    if is_confirm:
        print("(%s) %s" % (cwd, to_str()))
        if not query_yes_no("confirm?"):
            return False
        else:
            print("")

    return True


def _is_echo(command_line_args):
    return getattr(command_line_args, "echo", False)


def _is_confirm(command_line_args):
    return (
        getattr(command_line_args, "confirm", False)
        or _count_confirm_in_argv()
        or os.environ.get("__DODO_UNIVERSAL_CONFIRM__", None) == "1"
    )


def _is_echo(command_line_args):
    return getattr(command_line_args, "echo", False)


class Decorator:  # noqa
    def is_used(self, config, command_name, decorator_name):
        return True

    def add_arguments(self, parser):  # noqa
        parser.add_argument(
            "-c",
            "--confirm",
            action="count",
            help="Confirm each performed action before its execution."
            + " Use twice (e.g. dodo foo -cc) to indicate that also nested calls must be confirmed.",
        )

        parser.add_argument(
            "-e",
            "--echo",
            action="store_true",
            help="Print all commands instead of running them",
        )

    def modify_args(self, command_line_args, args_tree_root_node, cwd):  # noqa
        if _count_confirm_in_argv() > 1:
            local.env["__DODO_UNIVERSAL_CONFIRM__"] = "1"

        if _is_echo(command_line_args) and not Dodo.safe:
            raise CommandError(
                "The --echo option is not supported for unsafe commands.\n"
                + "Since this command is marked as unsafe, some operations will "
                + "be performed directly, instead of echoed to the console. "
                + "Use --confirm if you wish to execute with visual feedback. "
            )

        if _is_confirm(command_line_args) and not Dodo.safe:
            if not query_yes_no(
                "Since this command is marked as unsafe, some operations will "
                + "be performed without asking for confirmation. Continue?",
                default="no",
            ):
                sys.exit(1)

        if not _ask_confirmation(
            args_tree_root_node,
            cwd or local.cwd,
            _is_echo(command_line_args),
            _is_confirm(command_line_args),
        ):
            return None, None

        return args_tree_root_node, cwd



"""Module for working with the dodo configurations."""
import copy
import json
import os
import re
import sys

from dodo_commands.dependencies.get import dotenv, yaml
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.config_expander import ConfigExpander
from dodo_commands.framework.config_io import ConfigIO
from dodo_commands.framework.config_key import Key
from dodo_commands.framework.global_config import create_global_config  # noqa
from dodo_commands.framework.paths import Paths, _env

dotenv_values = dotenv.dotenv_values


def merge_into_config(config, layer, xpath=None):
    def _is_list(x):
        return isinstance(x, type(list()))

    def _is_dict(x):
        return isinstance(x, type(dict()))

    def _raise(xpath):
        raise CommandError(
            "Cannot merge configurations. Check key /%s" % "/".join(new_xpath)
        )

    xpath = xpath or []
    for key, val in (layer or {}).items():
        new_xpath = xpath + [key]

        if key not in config:
            config[key] = val
        elif _is_dict(val):
            if not _is_dict(config[key]):
                _raise(new_xpath)
            merge_into_config(config[key], val, new_xpath)
        elif _is_list(val):
            if not _is_list(config[key]):
                _raise(new_xpath)
            config[key].extend(val)
        else:
            config[key] = val


def _add_to_config(config, section, key, value):
    if section in config:
        if key not in config[section]:
            config[section][key] = value


def _system_commands_dir():
    """Return directory where system command scripts are stored"""
    import dodo_commands.dodo_system_commands

    return os.path.dirname(dodo_commands.dodo_system_commands.__file__)


def _extend_config(config):
    """Add special values to the project's config"""
    project_dir = Paths().project_dir()
    if project_dir:
        _add_to_config(config, "ROOT", "env_name", _env())
        _add_to_config(config, "ROOT", "project_dir", project_dir)
        _add_to_config(config, "ROOT", "config_dir", Paths().config_dir())


def extend_command_path(config):
    """Add the system commands to the command path"""
    extended_config = copy.deepcopy(config)
    _add_to_config(extended_config, "ROOT", "command_path", [])
    extended_config["ROOT"]["command_path"].append(_system_commands_dir())
    if not Paths().project_dir():
        extended_config["ROOT"]["command_path"].append(
            os.path.join(Paths().default_commands_dir(), "*")
        )
    return extended_config


def _report(x):
    sys.stderr.write(x)
    sys.stderr.flush()


def check_conflicts(layer_paths):
    generic_paths = {}
    for path in layer_paths:
        parts = os.path.basename(path).split(".")
        if len(parts) == 3:
            generic_path = os.path.join(os.path.dirname(path), parts[0])

            conflicting_path = generic_paths.get(generic_path, None)
            if conflicting_path:
                raise CommandError(
                    "Conflicting layers: %s and %s" % (path, conflicting_path)
                )

            generic_paths[generic_path] = path

    return layer_paths


def build_config(layers):
    config = {"ROOT": {}}
    for layer in layers:
        merge_into_config(config, layer)

    _extend_config(config)
    extra_vars = dict()

    def _load_env(dotenv_file):
        if not os.path.exists(dotenv_file):
            _report("Dotenv file not found: %s\n" % dotenv_file)
        extra_vars.update(dotenv_values(dotenv_file))

    # Call dotenv_values for every item of /ENV/dotenv
    callbacks = {}
    for idx, _ in enumerate(config["ROOT"].get("dotenv_files", [])):
        callbacks["/ROOT/dotenv_files/%d" % idx] = _load_env

    warnings = ConfigExpander(extra_vars).run(config, callbacks=callbacks)
    return (config, warnings)


def load_config(layer_filenames, config_io=None):
    layers = []
    config_io = config_io or ConfigIO()
    try:
        for layer_filename in layer_filenames:
            layers.append(config_io.load(layer_filename))
    except yaml.scanner.ScannerError:
        _report(
            "There was an error while loading the configuration. "
            "Run 'dodo diff' to compare your configuration to the "
            "default one.\n"
        )

    config, warnings = build_config(layers)
    return config


def expand_keys(config, text):
    result = ""
    val_terms = re.split("\$\{([^\}]+)\}", text)
    for idx, term in enumerate(val_terms):
        if idx % 2:
            str_rep = json.dumps(Key(config, term).get())
            if str_rep.startswith('"') and str_rep.endswith('"'):
                str_rep = str_rep[1:-1]
            result += str_rep
        else:
            result += term
    return result



import os

from dodo_commands.dependencies.get import six
from dodo_commands.dodo_system_commands._create_env_dir import create_env_dir
from dodo_commands.framework.paths import Paths
from dodo_commands.framework.util import symlink

configparser = six.moves.configparser


def global_config_get(config, section, key, default=""):
    if config is None:
        config = load_global_config_parser()
    return config.get(section, key) if config.has_option(section, key) else default


def load_global_config_parser():
    config_parser = configparser.ConfigParser()
    config_parser.read(Paths().global_config_filename())
    return config_parser


def write_global_config_parser(config_parser):
    """Save configuration."""
    with open(Paths().global_config_filename(), "w") as f:
        config_parser.write(f)


def projects_dir():
    return os.path.expanduser(
        load_global_config_parser().get("settings", "projects_dir")
    )


_global_config = """[settings]
projects_dir=~/projects
python_interpreter=python
diff_tool=diff

[recent]
"""

_default_config = """ROOT:
  version: 1.0.0
  command_path:
  - ${/ROOT/project_dir}/commands/*
"""


def _touch_init_py(dir_name):
    init_py = os.path.join(dir_name, "__init__.py")
    if not os.path.exists(init_py):
        with open(init_py, "w"):
            pass


def create_global_config():
    """Create config file and default_commands dir."""
    base_dir = Paths().global_config_dir()
    if not os.path.exists(base_dir):
        os.mkdir(base_dir)

    config_filename = Paths().global_config_filename()
    if not os.path.exists(config_filename):
        with open(config_filename, "w") as f:
            f.write(_global_config)

    default_project_dir = Paths().default_project_dir()

    if not os.path.exists(default_project_dir):
        os.makedirs(default_project_dir)
        default_config_filename = os.path.join(default_project_dir, "config.yaml")
        with open(default_config_filename, "w") as f:
            f.write(_default_config)

    default_env_dir = Paths().env_dir("default")
    if not os.path.exists(default_env_dir):
        create_env_dir(
            "default", default_env_dir, default_project_dir, default_project_dir, None
        )

    preset_commands_dirs = ("dodo_standard_commands", "dodo_docker_commands")
    global_commands_dir = Paths().global_commands_dir()
    if not os.path.exists(global_commands_dir):
        os.mkdir(global_commands_dir)
        _touch_init_py(global_commands_dir)
        for d in preset_commands_dirs:
            symlink(
                os.path.join(Paths().extra_dir(), d),
                os.path.join(global_commands_dir, d),
            )

    default_commands_dir = Paths().default_commands_dir()
    if not os.path.exists(default_commands_dir):
        os.makedirs(default_commands_dir)
        _touch_init_py(default_commands_dir)
        for d in preset_commands_dirs:
            symlink(
                os.path.join(global_commands_dir, d),
                os.path.join(default_commands_dir, d),
            )



import os

from dodo_commands import CommandError, Dodo
from dodo_commands.framework.command_path import get_command_dirs_from_config
from dodo_commands.framework.global_config import load_global_config_parser
from dodo_commands.framework.util import chop


def _args():
    Dodo.parser.add_argument("package", choices=_drop_dir_by_package_name().keys())
    return Dodo.parse_args()


def _diff_tool():
    return load_global_config_parser().get("settings", "diff_tool")


def _read_dot_drop_path(path):
    with open(path) as ifs:
        base_dir = os.path.realpath(os.path.dirname(path))
        rel_dir = chop(ifs.read())
        return os.path.realpath(os.path.join(base_dir, rel_dir))


def _register_drop_dir(drop_dir_by_package_name, package_name, drop_src_dir):
    if package_name in drop_dir_by_package_name:
        if drop_src_dir == drop_dir_by_package_name[package_name]:
            return
        else:
            print(
                "Warning: Found two drop directories for package %s:\n" % package_name
            )
            print(drop_src_dir)
            print(drop_dir_by_package_name[package_name])
    drop_dir_by_package_name[package_name] = drop_src_dir


def _dot_drop_path(item):
    result = os.path.join(item, ".drop-in")
    return result if os.path.exists(result) else None


def _drop_src_dir(item):
    result = os.path.join(item, "drop-in")
    return result if os.path.isdir(result) else None


def _drop_dir_by_package_name():
    drop_dir_by_package_name = {}
    command_dirs = get_command_dirs_from_config(Dodo.get())
    for item in command_dirs:
        dot_drop_path = _dot_drop_path(item)
        default_drop_src_dir = _drop_src_dir(item)

        if not dot_drop_path and not default_drop_src_dir:
            continue

        if dot_drop_path and default_drop_src_dir:
            print("Warning: found two drop-ins where only one was expected:\n")
            print("%s\n%s" % (dot_drop_path, default_drop_src_dir))

        package_name = os.path.basename(item)
        drop_src_dir = default_drop_src_dir or _read_dot_drop_path(dot_drop_path)
        _register_drop_dir(drop_dir_by_package_name, package_name, drop_src_dir)

    return drop_dir_by_package_name


if Dodo.is_main(__name__, safe=True):
    args = _args()

    drop_src_dir = _drop_dir_by_package_name().get(args.package)
    if not drop_src_dir:
        raise CommandError("No drop-in found for package %s" % args.package)

    if not os.path.isdir(drop_src_dir):
        raise CommandError("The drop-in directory does not exist: %s" % drop_src_dir)

    config_dir = Dodo.get("/ROOT/config_dir")
    drops_target_dir = os.path.join(config_dir, "drops")
    if not os.path.exists(drops_target_dir):
        Dodo.run(["mkdir", drops_target_dir])
    target_dir = os.path.join(drops_target_dir, args.package)

    if os.path.exists(target_dir):
        msg = ""
        msg += "Target directory already exists: %s\n" % target_dir
        msg += "\nThis probably means that %s has been already dropped into %s.\n" % (
            args.package,
            config_dir,
        )
        msg += "To update the drop-in manually, please run:\n\n"
        msg += "%s %s %s" % (_diff_tool(), drop_src_dir, target_dir)
        raise CommandError(msg)

    Dodo.run(["cp", "-rf", drop_src_dir, target_dir])

    readme_path = os.path.join(drop_src_dir, "README.md")
    if os.path.exists(readme_path):
        Dodo.run(["cat", readme_path])



import sys
from argparse import ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import six
from dodo_commands.framework import ramda as R
from dodo_commands.framework.config_io import ConfigIO
from dodo_commands.framework.config_key import Key
from dodo_commands.framework.global_config import load_global_config_parser
from dodo_commands.framework.paths import Paths

configparser = six.moves.configparser


def _args():
    parser = ArgumentParser(description=("Edit the dodo configuration files"))

    parser.add_argument(
        "--key", help="Only edit this key. Used in conjunction with --val"
    )
    parser.add_argument(
        "--val", help="The value to be used in conjunction with the --key option"
    )

    args = Dodo.parse_args(parser)

    if args.key or args.val:
        if not args.key and args.val:
            raise CommandError(
                "The options --key and --val should always be used together"
            )

    args.editor = load_global_config_parser().get("settings", "config_editor")
    args.config_dir = Dodo.get("/ROOT/config_dir")
    return args


if Dodo.is_main(__name__, safe=("--key" not in sys.argv)):
    try:
        args = _args()
    except configparser.NoOptionError as e:
        raise CommandError(
            "{error}. Please check {filename}".format(
                error=str(e), filename=Paths().global_config_filename()
            )
        )

    config = ConfigIO().load()

    if args.key and args.val:
        key = Key(config, args.key)
        key.set(args.val)
        ConfigIO().save(config)
        sys.exit(0)

    def add_global_config_filename(layer_paths):
        return R.concat(layer_paths, [Paths().global_config_filename()])

    x = Dodo.get_container().layers.get_ordered_layer_paths()
    x = add_global_config_filename(x)
    yaml_filenames = x

    Dodo.run(args.editor.split() + yaml_filenames, cwd=".")



from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.util import xpath_to_string


class KeyNotFound(CommandError):
    """Raised if an xpath is not found in a configuration."""

    pass


class Key:
    """Access a nested value in a dict."""

    def __init__(self, config, xpath_or_str):  # noqa
        self.config = config

        def split_on_slash(key):
            return [k for k in key[1:].split("/") if k]

        # the list of sub-keys along the path to an item in self.config
        self.xpath = (
            split_on_slash(xpath_or_str)
            if isinstance(xpath_or_str, str)
            else xpath_or_str
        )

    def _step_into(self, node, subkey):
        try:
            if isinstance(node, type(dict())):
                return node[subkey]
            if isinstance(node, type(list())):
                return node[int(subkey)]
        except:  # noqa
            raise KeyNotFound("Cannot get value at %s\n" % self)

    def exists(self):
        config_node = self.config
        for xpath_node in self.xpath:
            if isinstance(config_node, dict) and xpath_node in config_node:
                config_node = config_node[xpath_node]
            elif isinstance(config_node, list) and int(xpath_node) < len(config_node):
                config_node = config_node[int(xpath_node)]
            else:
                return False
        return True

    def get(self):  # noqa
        node = self.config
        for subkey in self.xpath:
            node = self._step_into(node, subkey)
        return node

    def set(self, value):  # noqa
        node = self.config
        for subkey in self.xpath[:-1]:
            node = self._step_into(node, subkey)
        node[self.xpath[-1]] = value

    def remove(self):  # noqa
        node = self.config
        for subkey in self.xpath[:-1]:
            node = self._step_into(node, subkey)
        del node[self.xpath[-1]]

    def __repr__(self):  # noqa
        return xpath_to_string(self.xpath)



import os
import re
import subprocess
import sys
from importlib import import_module

from dodo_commands.dependencies import yaml_round_trip_load
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.command_map import CommandMapItem
from dodo_commands.framework.command_path import extend_sys_path
from dodo_commands.framework.config import Paths
from dodo_commands.framework.util import query_yes_no

CommandNotFound = plumbum.commands.processes.CommandNotFound


class PythonCommandMapItem(CommandMapItem):
    def __init__(self, group, filename):
        super(PythonCommandMapItem, self).__init__(group, filename, "py")
        self.package_path = group


class PythonCommandHandler:
    def add_commands_to_map(self, command_dirs, file_map, command_map):
        extend_sys_path(command_dirs)
        for command_dir, files in file_map.items():
            for file in files:
                command_name, ext = os.path.splitext(os.path.basename(file))
                if ext == ".py" and not command_name.startswith("_"):
                    command_map[command_name] = PythonCommandMapItem(
                        group=os.path.basename(command_dir), filename=file
                    )

    def execute(self, command_map_item, command_name):
        def meta_data_filename():
            base_path = os.path.dirname(command_map_item.filename)
            return os.path.join(base_path, command_name + ".meta")

        def meta_data():
            if os.path.exists(meta_data_filename()):
                with open(meta_data_filename()) as f:
                    return yaml_round_trip_load(f.read())
            return None

        def install_python_packages(meta_data, dependency):
            """Pip install packages found in meta_data."""
            requirements = [
                x
                for x in meta_data["requirements"]
                if re.search(r"\b" + dependency + r"\b", x)
            ]
            if requirements:
                if len(requirements) > 1:
                    raise CommandError(
                        "Found more than 1 candidate for python package %s in %s"
                        % (dependency, meta_data_filename())
                    )

                requirement = requirements[0]
                print("This command wants to install %s:\n" % requirement)
                if query_yes_no("Install (yes), or abort (no)?"):
                    subprocess.check_call(
                        [sys.executable, "-m", "pip", "install", requirement]
                    )
                    print("--- Done ---\n\n")

        import_path = "%s.%s" % (command_map_item.package_path, command_name)
        if command_map_item.package_path in ("", None, "."):
            import_path = command_name

        # Try the import, return if success
        try:
            return import_module(import_path)
        except ImportError as e:
            # Try the import again, but first run install_python_packages
            dependency = (e.name,)
            try:
                md = meta_data()
                if md:
                    install_python_packages(md, dependency)
                import_module(import_path)
            except ImportError:
                raise CommandError("Could not import python module: %s" % dependency)



from dodo_commands.framework import ramda as R

from .command_line import CommandLine  # noqa
from .commands import Commands, init_commands  # noqa
from .config import Config  # noqa
from .layers import Layers, init_layers  # noqa


def i_(facet_class, member, prefix=None, alt_name=None):
    arg_name = alt_name if alt_name else prefix + "_" + member if prefix else member
    return ("in", facet_class, member, arg_name)


def o_(facet_class, member):
    return ("out", facet_class, member)


def map_datas(*args, transform):
    def action(ctr):
        kwargs = {}

        def is_input(arg):
            return arg[0] == "in"

        def do_add_to_kwargs(input_arg):
            _, facet_class, member, arg_name = input_arg
            value = getattr(facet_class.get(ctr), member)
            kwargs[arg_name] = value

        x = args
        # [(in | out, facet_class, member)]
        x = R.filter(is_input)(x)
        # [(in | out, facet_class, member)]
        x = R.for_each(do_add_to_kwargs)(x)
        # [value]
        output_values = transform(**kwargs)

        def is_output(arg):
            return arg[0] == "out"

        def zip_with_output_values(output_args):
            return [
                [output_arg, output_values[output_arg[2]]] for output_arg in output_args
            ]

        def do_store(output_arg, output_value):
            _, facet_class, member = output_arg
            setattr(facet_class.get(ctr), member, output_value)

        x = args
        # [(in | out, facet_class, member)]
        x = R.filter(is_output)(x)
        # [(in | out, facet_class, member)]
        x = zip_with_output_values(x)
        # [((in | out, facet_class, member), output_value)]
        R.for_each(R.ds(do_store))(x)

        return output_values

    return action


def register(*args):
    def decorator(func):
        setattr(func, "action_args", args)
        return func

    return decorator


def run(ctr, action):
    return map_datas(
        *action.action_args,
        transform=action,
    )(ctr)



import re
from argparse import ArgumentParser

from dodo_commands.dependencies import yaml_round_trip_dump
from dodo_commands.framework.singleton import Dodo


def _args():
    parser = ArgumentParser(description="Print the full configuration.")
    parser.add_argument("key", nargs="?")
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__):
    args = _args()
    if args.key:
        prefix = "" if args.key.startswith("/") else "/"
        contents = Dodo.get(prefix + args.key)
        if isinstance(contents, str):
            print(contents)
        else:
            print("%s" % yaml_round_trip_dump(contents))
    else:
        content = re.sub(
            r"^([0-9_A-Z]+\:)$",
            r"\n\1",
            yaml_round_trip_dump(Dodo.get()),
            flags=re.MULTILINE,
        )
        print(re.sub(r"^\n\n", r"\n", content, flags=re.MULTILINE))



import os

from dodo_commands import Dodo


def _args():
    Dodo.parser.add_argument(
        "container_type", choices=Dodo.get("/DOCKER/containers").keys()
    )
    Dodo.parser.add_argument("output_dir")
    Dodo.parser.add_argument("--reverse", action="store_true")
    return Dodo.parse_args()


def _docker_run(output_dir, args):
    Dodo.run(
        [
            "docker",
            "run",
            "--rm",
            "--volumes-from",
            container_name,
            "--volume",
            "%s:/tmp/docker-snapshot" % output_dir,
            container_type["image"],
        ]
        + args,
        cwd=".",
    )


if Dodo.is_main(__name__, safe=True):
    args = _args()
    container_name = Dodo.get("/DOCKER/containers/" + args.container_type)

    docker_options = Dodo.get("/DOCKER_OPTIONS", {})
    docker_options.setdefault("docker-snapshot", {})
    docker_options["docker-snapshot"]["volumes_from"] = container_name

    container_type = Dodo.get("/DOCKER/container_types/" + args.container_type)
    for path in container_type["dirs"]:
        host_output_path = os.path.join(args.output_dir, path[1:])
        docker_output_path = os.path.join("/tmp/docker-snapshot", path[1:])

        if args.reverse:
            _docker_run(args.output_dir, ["rm", "-rf", os.path.join(path, "*")])
            _docker_run(
                args.output_dir,
                ["cp", "-rf", os.path.join(docker_output_path, "*"), path],
            )
        else:
            if os.path.exists(host_output_path):
                Dodo.run(["rm", "-rf", host_output_path], cwd=".")
            Dodo.run(["mkdir", "-p", os.path.dirname(host_output_path)], cwd=".")
            _docker_run(
                args.output_dir,
                ["cp", "-rf", path, os.path.dirname(docker_output_path)],
            )



import os

from dodo_commands.framework import ramda as R
from dodo_commands.framework.config_io import ConfigIO


class Layers:
    def __init__(self):
        self.config_io = ConfigIO()
        self.root_layer_path = None
        self.root_layer = None
        self.layer_by_target_path = {}
        self.selected_layer_by_path = {}
        self.metadata_by_layer_name = None

    def get_ordered_layer_paths(self):
        root_layer_path = self.config_io.glob([self.root_layer_path])[0]

        x = R.concat(
            self.selected_layer_by_path.keys(), self.layer_by_target_path.keys()
        )
        x = R.uniq(x)
        x = sorted(x, key=os.path.basename)
        x = self.config_io.glob(x)

        x = R.filter(lambda x: x != root_layer_path)(x)
        x = R.concat([root_layer_path], x)

        return x

    @staticmethod
    def get(ctr):
        return ctr.layers


def init_layers(self, root_layer_path):
    self.root_layer_path = root_layer_path
    return self



from argparse import ArgumentParser

from dodo_commands import DecoratorScope, Dodo

from dodo_docker_commands.decorators.docker import Decorator as DockerDecorator


def _choices():
    choices = []
    for key in Dodo.get("/DOCKER_OPTIONS", {}).keys():
        keys = [key] if isinstance(key, str) else key
        for x in keys:
            if x not in choices and not x.startswith("!"):
                choices.append(str(x))
    return choices


def _args():
    parser = ArgumentParser(description="Opens a shell in the docker container.")
    parser.add_argument(
        "service",
        choices=_choices(),
        help=("Use this key to look up the docker options in /DOCKER_OPTIONS"),
    )
    parser.add_argument(
        "--image",
        choices=Dodo.get("/DOCKER_IMAGES", {}).keys(),
        help=("Use the docker image stored under this key in /DOCKER_IMAGES"),
    )
    parser.add_argument("--image-name", help=("Use the docker image with this name"))
    parser.add_argument(
        "--name", help=("Override the name of the started docker container")
    )
    parser.add_argument("--command")
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__):
    args = _args()

    docker_options = DockerDecorator.merged_options(Dodo.get, args.service)

    if args.image:
        docker_options["image"] = Dodo.get(
            "/DOCKER_IMAGES/%s/image" % args.image, args.image
        )
    elif args.image_name:
        docker_options["image"] = args.image_name

    if args.name:
        docker_options["name"] = args.name
    else:
        docker_options["name"] = args.service

    Dodo.get()["DOCKER_OPTIONS"] = {Dodo.command_name: docker_options}

    with DecoratorScope("docker"):
        Dodo.run(
            [args.command] if args.command else ["sh"],
            cwd=docker_options.get("cwd", "/"),
        )



import os
import shutil
import subprocess
import sys

from setuptools import find_packages, setup
from setuptools.command.install import install


class InstallPrivatePackages(install):
    def _package_dirname(self):
        return os.path.join(
            self.install_lib, "dodo_commands", "dependencies", "packages"
        )

    def _bin_dirname(self):
        return os.path.join(self.install_lib, "dodo_commands", "bin")

    def _install_packages(self, package_dirname):
        if not os.path.exists(package_dirname):
            os.makedirs(package_dirname)

        for dependency in [
            "python-dotenv==0.12.0",
            "plumbum==1.6.8",
            "ruamel.yaml==0.16.10",
            "parsimonious==0.8.1",
            "six==1.14.0",
            "funcy==1.14",
            "ansimarkup==1.4.0",
            "argcomplete==1.11.1",
            "semantic_version==2.8.4",
        ]:
            try:
                subprocess.check_call(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        "--target",
                        package_dirname,
                        dependency,
                    ]
                )
            except:  # noqa
                pass

    def _install_autocomplete_scripts(self, bin_dirname):
        bash_autocomplete_dir = "/etc/bash_completion.d/"
        fish_autocomplete_dir = "/etc/fish/conf.d/"

        for (f, d) in (
            ("dodo_autocomplete.sh", bash_autocomplete_dir),
            ("dodo_autocomplete.fish", fish_autocomplete_dir),
        ):
            try:
                if not os.path.exists(d):
                    os.makedirs(d)
                shutil.copy(os.path.join(bin_dirname, f), d)
            except:  # noqa
                pass

    def run(self):
        super(InstallPrivatePackages, self).run()
        self._install_packages(self._package_dirname())
        self._install_autocomplete_scripts(self._bin_dirname())


setup(
    name="dodo_commands",
    version="0.36.0",
    description="Project-aware development environments, inspired by django-manage",
    url="https://github.com/mnieber/dodo_commands",
    download_url="https://github.com/mnieber/dodo_commands/tarball/0.36.0",
    author="Maarten Nieber",
    author_email="hallomaarten@yahoo.com",
    license="MIT",
    packages=find_packages(),
    package_data={
        "dodo_commands": [
            "bin/*.sh",
            "bin/*.fish",
            "extra/dodo_standard_commands/*.meta",
            "extra/dodo_docker_commands/*.meta",
            "extra/fish/functions/*.fish",
            "extra/fish/conf.d/*.fish",
        ]
    },
    entry_points={
        "console_scripts": [
            "dodo=dodo_commands.dodo:main",
        ]
    },
    data_files=[
        (
            "/etc/bash_completion.d",
            [
                "dodo_commands/bin/dodo_autocomplete.sh",
                "dodo_commands/bin/sdodo_autocomplete.sh",
            ],
        ),
        ("/etc/fish/conf.d", ["dodo_commands/bin/dodo_autocomplete.fish"]),
    ],
    cmdclass={"install": InstallPrivatePackages},
    install_requires=[],
    zip_safe=False,
)



"""Decorates command lines with docker arguments."""
import os
from fnmatch import fnmatch

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.args_tree import ArgsTreeNode
from dodo_commands.framework.config import merge_into_config
from dodo_commands.framework.decorator_utils import uses_decorator

local = plumbum.local


def invert_path(path):
    def _len(path):
        return len(path.split(os.sep))

    options = Decorator.merged_options(Dodo.get, Dodo.command_name)
    inverse_volume_map = options.get("inverse_volume_map", {})
    best_common_path = ""
    best_host_path = ""

    for container_path, host_path in inverse_volume_map.items():
        common_path = os.path.commonpath([path, container_path])
        if _len(common_path) > _len(best_common_path):
            best_common_path = common_path
            best_host_path = os.path.join(host_path, os.path.relpath(path, common_path))

    return best_host_path


def _is_tuple(x):
    return hasattr(x, "__len__") and not isinstance(x, type(str()))


# Resp: reads docker related settings from the docker_config and uses
# them to modify the current args_tree_root_node.


class Decorator:  # noqa
    def is_used(self, config, command_name, decorator_name):
        return uses_decorator(config, command_name, decorator_name)

    @classmethod
    def _add_docker_volume_list(cls, docker_config, root_node):
        volume_list = docker_config.get("volume_list", [])
        volume_map = docker_config.get("volume_map", {})

        volume_map_strict = docker_config.get("volume_map_strict", {})
        for key, val in volume_map_strict.items():
            if not os.path.exists(key):
                raise CommandError("Path in volume_map_strict not found: %s" % key)
            volume_map[key] = val

        options = ["--volume=%s:%s" % (x, x) for x in volume_list] + [
            "--volume=%s:%s" % key_val for key_val in volume_map.items()
        ]
        for x in options:
            root_node["volume"].append(x)

    @classmethod
    def _add_docker_publish_list(cls, docker_config, root_node):
        publish_list = docker_config.get("publish_list", [])
        publish_map = docker_config.get("publish_map", {})
        options = ["--publish=%s:%s" % (x, x) for x in publish_list] + [
            "--publish=%s:%s" % key_val for key_val in publish_map.items()
        ]
        for x in options:
            root_node["publish"].append(x)

    @classmethod
    def _add_docker_volumes_from_list(cls, docker_config, root_node):
        volumes_from_list = docker_config.get("volumes_from_list", [])
        options = ["--volumes-from=%s" % x for x in volumes_from_list]
        for x in options:
            root_node["volumes-from"].append(x)

    @classmethod
    def _add_docker_variable_list(cls, docker_config, root_node):
        variable_list = docker_config.get("variable_list", [])
        variable_map = docker_config.get("variable_map", {})
        options = ["--env=%s" % x for x in variable_list] + [
            "--env=%s=%s" % key_val for key_val in variable_map.items()
        ]
        for x in options:
            root_node["env"].append(x)

    @classmethod
    def _add_linked_container_list(cls, docker_config, root_node):
        options = ["--link=%s" % x for x in docker_config.get("link_list", [])]
        for x in options:
            root_node["link"].append(x)

    @classmethod
    def _add_interactive(cls, docker_config, docker_node):
        if docker_config.get("is_interactive", True):
            docker_node["basic"].append("--interactive")
            docker_node["basic"].append("--tty")

    @classmethod
    def _add_workdir(cls, docker_config, docker_node, cwd):
        if cwd and not ("workdir" in docker_config and not docker_config["workdir"]):
            docker_node["workdir"].append("--workdir=" + cwd)
        elif docker_config.get("workdir", ""):
            docker_node["workdir"].append("--workdir=" + docker_config.get("workdir"))

    @classmethod
    def _add_user(cls, docker_config, docker_node):
        if docker_config.get("user", False):
            docker_node["basic"].extend(["--user", docker_config["user"]])

    @classmethod
    def _add_environment_vars(cls, get_config, docker_config, docker_node):
        for key_val in get_config("/ENVIRONMENT/variable_map", {}).items():
            docker_node["env"].append("--env=%s=%s" % key_val)

    @classmethod
    def _add_extra_options(cls, docker_config, docker_node):
        for x in docker_config.get("extra_options", []):
            docker_node["other"].append(x)

    @classmethod
    def merged_options(cls, get_config, command_name):
        merged = {}
        docker_options = get_config("/DOCKER_OPTIONS", {})
        for patterns, docker_config in docker_options.items():
            should_merge = False
            for pattern in patterns if _is_tuple(patterns) else [patterns]:
                if pattern.startswith("!"):
                    should_merge = should_merge and not fnmatch(
                        command_name, pattern[1:]
                    )
                else:
                    should_merge = should_merge or fnmatch(command_name, pattern)
            if should_merge:
                merge_into_config(merged, docker_config)
        return merged

    @classmethod
    def docker_node(cls, get_config, command_name, cwd):
        merged = cls.merged_options(get_config, command_name)

        image, container = merged.get("image"), merged.get("container")

        if image:
            docker_node = ArgsTreeNode("docker", args=["docker", "run"])
            docker_node.add_child(ArgsTreeNode("name"))
            docker_node.add_child(ArgsTreeNode("basic"))
            docker_node.add_child(ArgsTreeNode("link"))
            docker_node.add_child(ArgsTreeNode("publish"))
            docker_node.add_child(ArgsTreeNode("volume", is_horizontal=False))
            docker_node.add_child(ArgsTreeNode("volumes-from"))
            docker_node.add_child(ArgsTreeNode("env", is_horizontal=False))
            docker_node.add_child(ArgsTreeNode("other"))
            docker_node.add_child(ArgsTreeNode("workdir"))
            docker_node.add_child(ArgsTreeNode("image"))

            cls._add_linked_container_list(merged, docker_node)
            cls._add_docker_publish_list(merged, docker_node)
            cls._add_docker_volume_list(merged, docker_node)
            cls._add_docker_volumes_from_list(merged, docker_node)
            cls._add_docker_variable_list(merged, docker_node)
            cls._add_workdir(merged, docker_node, cwd)
            cls._add_interactive(merged, docker_node)
            cls._add_user(merged, docker_node)
            cls._add_environment_vars(get_config, merged, docker_node)
            cls._add_extra_options(merged, docker_node)

            if merged.get("rm", True):
                docker_node["basic"].append("--rm")

            if merged.get("is_daemon", False):
                docker_node["basic"].append("-d")

            name = merged.get("name", command_name)
            if name:
                docker_node["name"].append("--name=%s" % name)

            docker_node["image"].append(image)
        elif container:
            docker_node = ArgsTreeNode("docker", args=["docker", "exec"])
            docker_node.add_child(ArgsTreeNode("basic"))
            docker_node.add_child(ArgsTreeNode("env", is_horizontal=False))
            docker_node.add_child(ArgsTreeNode("other"))
            docker_node.add_child(ArgsTreeNode("workdir"))
            docker_node.add_child(ArgsTreeNode("container"))

            cls._add_interactive(merged, docker_node)
            cls._add_docker_variable_list(merged, docker_node)
            cls._add_workdir(merged, docker_node, cwd)
            cls._add_user(merged, docker_node)
            cls._add_environment_vars(get_config, merged, docker_node)
            cls._add_extra_options(merged, docker_node)
            docker_node["container"].append(container)
        else:
            raise CommandError(
                "No docker image or container found in /DOCKER_OPTIONS for command %s"
                % command_name
            )

        return docker_node, image

    def add_arguments(self, parser):  # noqa
        pass

    def modify_args(self, command_line_args, args_tree_root_node, cwd):  # override
        docker_node, _ = self.docker_node(Dodo.get_config, Dodo.command_name, cwd)

        docker_node.add_child(args_tree_root_node)
        return docker_node, local.cwd



import os
import sys
import time
from argparse import SUPPRESS, ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.choice_picker import ChoicePicker
from dodo_commands.framework.util import exe_exists

tmux = plumbum.cmd.tmux


def _normalize(category):
    return category.replace(" ", "-")


def _args():
    command_map = Dodo.get("/MENU/commands", {})
    default_session_id = Dodo.get("/MENU/session_id", os.path.expandvars("$USER"))

    parser = ArgumentParser()
    parser.add_argument(
        "category",
        choices=["all"] + list([_normalize(x) for x in command_map.keys()]),
        nargs="?",
    )
    parser.add_argument("--contnue", action="store_true", help=SUPPRESS)
    parser.add_argument("--tmux", action="store_true")
    parser.add_argument("--list", action="store_true")
    parser.add_argument("--run", type=int, nargs="?", const=-1)
    parser.add_argument(
        "--id",
        dest="session_id",
        default=default_session_id,
        help="The tmux session id",
    )

    args = Dodo.parse_args(parser)
    args.category = args.category or "all"
    args.run = -1 if args.run is None else args.run
    args.command_map = command_map

    return args


def _create_tmux_window(session_id):
    # Create tmux session
    tmux("-2", "new-session", "-d", "-s", session_id)

    # Create a tmux window
    tmux("new-window", "-t", "%s:1" % session_id, "-n", "Logs")


def _get_categories(command_map, category):
    return [x for x in command_map if category == "all" or _normalize(x) == category]


def _get_commands_and_labels(command_map, category):
    categories = _get_categories(command_map, category)
    label_size = 0
    for category in categories:
        label_size = max(label_size, len(category))
    label_prefix = "%0" + str(label_size) + "s"

    commands, labels = [], []
    for category in categories:
        for command in command_map[category]:
            commands.append(command)
            format_string = "%02s [" + label_prefix + "] - %s"
            labels.append(format_string % (str(len(commands)), category, command))

    return commands, labels


def _get_selected_commands(commands, labels, allow_free_text=False):
    class Picker(ChoicePicker):
        def print_choices(self, choices):
            print()
            for idx, label in enumerate(choices):
                print(label)
            print()

        def question(self):
            return "Select one or more commands (e.g. 1,3-4)%s or type 0 to exit: " % (
                ", or type a command," if allow_free_text else ""
            )

        def on_invalid_index(self, index):
            if index == 0:
                sys.exit(0)

    picker = Picker(commands, allow_free_text=allow_free_text, labels=labels)
    picker.pick()
    return [picker.free_text] if picker.free_text else picker.get_choices()


if Dodo.is_main(__name__):
    args = _args()
    commands, labels = _get_commands_and_labels(args.command_map, args.category)

    if not commands:
        raise CommandError("No commands were found in the /MENU configuration key")

    if args.list:
        print()
        for label in labels:
            print(label)
    elif args.tmux:
        if not exe_exists("tmux"):
            raise CommandError("Tmux is not installed on this sytem.")

        has_session = False
        try:
            sessions = tmux("ls")
            for session in sessions.split("\n"):
                has_session = has_session or session.startswith("%s:" % args.session_id)
        except:  # noqa
            pass

        if not has_session:
            _create_tmux_window(args.session_id)
            tmux("send-keys", " ".join(sys.argv + ["--contnue"]), "C-m")
            # Attach to tmux session
            # HACK: why does this only work via Dodo.run?
            Dodo.run(
                ["tmux", "-2", "attach-session", "-t", args.session_id],
            )
        elif not args.contnue:
            Dodo.run(
                ["tmux", "-2", "attach-session", "-t", args.session_id],
            )
        else:
            while True:
                tmux("send-keys", "-R", "")
                selected_commands = _get_selected_commands(
                    commands, labels, allow_free_text=True
                )
                for command in selected_commands:
                    print(command)
                    tmux("split-window", "-v")
                    time.sleep(0.5)
                    tmux("send-keys", command, "C-m")
                    tmux("select-layout", "tile")

                # Set default window
                tmux("select-pane", "-t", "0")
    else:
        selected_commands = (
            _get_selected_commands(commands, labels)
            if args.run == -1
            else [commands[args.run - 1]]
        )
        for command in selected_commands:
            Dodo.run(["bash", "-c", command])



from dodo_commands.framework import ramda as R
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.get_aliases import get_aliases


def get_inferred_command_map(
    global_aliases,
    metadata_by_layer_name,
    layer_by_target_path,
):
    layer_name_by_inferred_command = {}

    for layer_name, layer_metadata in metadata_by_layer_name.items():

        def get_layer(layer_metadata):
            return layer_by_target_path[layer_metadata.target_path]

        def get_cmd_aliases(layer):
            command_aliases = get_aliases(layer)
            for alias, target in global_aliases.items():
                if alias in command_aliases:
                    raise CommandError(
                        "Ambigous command alias %s, could be %s or %s"
                        % (alias, target, command_aliases[alias])
                    )
            return R.merge(command_aliases, global_aliases)

        def get_inferred_aliases(command_aliases, inferred_commands):
            return list(
                alias
                for alias in command_aliases.keys()
                if command_aliases[alias] in inferred_commands
            )

        def get_inferred_commands_and_aliases(inferred_commands, inferred_aliases):
            return R.uniq(inferred_commands + inferred_aliases)

        def do_store(inferred_command):
            if inferred_command in layer_name_by_inferred_command:
                raise CommandError(
                    "Ambigous inferred command %s. " % inferred_command
                    + "Could be %s or %s"
                    % (
                        layer_metadata.target_path,
                        layer_name_by_inferred_command[inferred_command],
                    )
                )
            layer_name_by_inferred_command[inferred_command] = layer_name

        x = layer_metadata
        x = layer = get_layer(x)
        x = command_aliases = get_cmd_aliases(layer)
        x = inferred_aliases = get_inferred_aliases(
            command_aliases, layer_metadata.inferred_commands
        )
        x = get_inferred_commands_and_aliases(
            layer_metadata.inferred_commands, inferred_aliases
        )
        R.for_each(do_store)(x)

    return layer_name_by_inferred_command



import os
from argparse import ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.framework.global_config import load_global_config_parser
from dodo_commands.framework.paths import Paths


def _args():
    parser = ArgumentParser(
        description="Show diff for a file in the Dodo Commands system directory."
    )
    parser.add_argument("file", nargs="?", help="Show diff for this file")
    parser.add_argument(
        "--env-name", help="Compare to files from an alternative environment"
    )
    args = Dodo.parse_args(parser)
    return args


def _diff_tool():
    return load_global_config_parser().get("settings", "diff_tool")


if Dodo.is_main(__name__):
    args = _args()
    file = args.file or "."

    project_dir = Paths().project_dir()
    if not project_dir:
        raise CommandError("No active dodo commands project")

    if args.env_name:
        ref_project_dir = os.path.abspath(
            os.path.join(project_dir, "..", args.env_name)
        )
        original_file = os.path.join(ref_project_dir, file)
        copied_file = os.path.join(project_dir, file)
    else:
        shared_config_dir = Dodo.get("/ROOT/shared_config_dir")
        original_file = os.path.realpath(os.path.join(shared_config_dir, file))
        copied_file = os.path.join(Paths().config_dir(), file)

    Dodo.run([_diff_tool(), original_file, copied_file])
